from copy import copy, deepcopy
import json
from tempfile import NamedTemporaryFile

import pytest
import responses

from datarobot import CustomInferenceModel, TARGET_TYPE
from datarobot.enums import CUSTOM_MODEL_TARGET_TYPE, CUSTOM_TASK_TYPE
from datarobot.utils import underscorize
from tests.model.utils import assert_custom_model_version


@pytest.fixture
def mocked_inference_model(mocked_models):
    return mocked_models["data"][1]


@pytest.fixture
def mocked_inference_model_with_resources(mocked_models_with_resources):
    return deepcopy(mocked_models_with_resources["data"][1])


@pytest.fixture
def mocked_multiclass_inference_model(mocked_multiclass_models):
    return deepcopy(mocked_multiclass_models["data"][1])


@pytest.fixture
def mocked_unstructured_inference_model(mocked_unstructured_models):
    return deepcopy(mocked_unstructured_models["data"][1])


@pytest.fixture
def mocked_anomaly_inference_model(mocked_anomaly_models):
    return deepcopy(mocked_anomaly_models["data"][1])


@pytest.fixture
def mocked_inference_model_no_latest_version(mocked_inference_model):
    model = copy(mocked_inference_model)
    model["latestVersion"] = None
    return model


@pytest.fixture
def mocked_inference_model_with_deprecated_target_support(mocked_inference_model):
    model = deepcopy(mocked_inference_model)
    if model["targetType"] == TARGET_TYPE.BINARY:
        model["supportsBinaryClassification"] = True
    elif model["targetType"] == TARGET_TYPE.REGRESSION:
        model["supportsRegression"] = True
    return model


def assert_custom_model_common(model, model_json, custom_task=False):
    assert model.id == model_json["id"]
    assert model.name == model_json["name"]
    assert model.description == model_json["description"]

    for expected_target_type in [
        TARGET_TYPE.BINARY,
        TARGET_TYPE.REGRESSION,
        TARGET_TYPE.MULTICLASS,
        TARGET_TYPE.UNSTRUCTURED,
        TARGET_TYPE.ANOMALY,
    ]:
        assert (model.target_type == expected_target_type) == (
            model_json["targetType"] == expected_target_type
        )

    if model_json["latestVersion"]:
        assert_custom_model_version(model.latest_version, model_json["latestVersion"])
    else:
        assert model.latest_version is None
    assert model.deployments_count == model_json["deploymentsCount"]
    assert model.created_by == model_json["createdBy"]
    assert model.updated_at == model_json["updated"]
    assert model.created_at == model_json["created"]

    if custom_task:
        if model.task_type == CUSTOM_TASK_TYPE.TRANSFORM:
            assert model_json["targetType"] == CUSTOM_TASK_TYPE.TRANSFORM


def assert_inference_model(model, model_json):
    assert_custom_model_common(model, model_json)
    assert model.language == model_json["language"]
    if model.target_type in CUSTOM_MODEL_TARGET_TYPE.REQUIRES_TARGET_NAME:
        assert model.target_name == model_json["targetName"]
    assert model.training_dataset_id == model_json["trainingDatasetId"]
    assert model.training_dataset_version_id == model_json["trainingDatasetVersionId"]
    assert model.training_data_file_name == model_json["trainingDataFileName"]
    assert (
        model.training_data_assignment_in_progress == model_json["trainingDataAssignmentInProgress"]
    )
    assert model.positive_class_label == model_json.get("positiveClassLabel")
    assert model.negative_class_label == model_json.get("negativeClassLabel")
    assert model.prediction_threshold == model_json.get("predictionThreshold")
    assert model.class_labels == model_json.get("classLabels")
    assert model.training_data_partition_column == model_json["trainingDataPartitionColumn"]
    assert model.network_egress_policy == model_json["networkEgressPolicy"]
    assert model.maximum_memory == model_json["maximumMemory"]
    assert model.replicas == model_json["replicas"]


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


class TestCustomInferenceModel(object):
    @pytest.mark.parametrize(
        "model_data", ["mocked_inference_model", "mocked_inference_model_no_latest_version"]
    )
    def test_from_server_data(self, request, model_data):
        model_data = request.getfixturevalue(model_data)
        model = CustomInferenceModel.from_server_data(model_data)
        assert_inference_model(model, model_data)

    @responses.activate
    def test_get_version(self, mocked_inference_model, make_models_url):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    def test_list_versions(self, make_models_url, mocked_inference_model):
        # arrange
        url = make_models_url() + "?customModelType=inference"
        mock_get_response(
            url,
            {
                "count": 1,
                "totalCount": 1,
                "next": None,
                "previous": None,
                "data": [mocked_inference_model],
            },
        )

        # act
        models = CustomInferenceModel.list()

        # assert
        assert len(models) == 1
        assert_inference_model(models[0], mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    @pytest.mark.parametrize("labels_as_files", [True, False])
    def test_update(self, make_models_url, mocked_inference_model, labels_as_files):
        # arrange
        patch = {
            "description": "xx",
            "predictionThreshold": 0.1,
            "classLabels": ["hot dog", "burrito", "hoagie", "reuben"],
        }

        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        mocked_inference_model.update(patch)

        responses.add(
            responses.PATCH,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        with NamedTemporaryFile() as f:
            args = deepcopy(patch)
            if labels_as_files:
                f.write("\n".join(args.pop("classLabels")).encode("utf-8"))
                f.flush()
                args["classLabelsFile"] = f.name

            model.update(**{underscorize(k): v for k, v in args.items()})

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url == url
        assert responses.calls[1].request.body == json.dumps(patch).encode()

    @responses.activate
    def test_refresh(self, make_models_url, mocked_inference_model):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)

        mocked_inference_model.update({"description": "xx", "predictionThreshold": 0.1})

        mock_get_response(url, mocked_inference_model)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.refresh()

        # assert
        assert_inference_model(model, mocked_inference_model)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_delete(self, make_models_url, mocked_inference_model):
        url = make_models_url(mocked_inference_model["id"])

        mock_get_response(url, mocked_inference_model)
        responses.add(
            responses.DELETE, url, status=204, content_type="application/json",
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.delete()

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "DELETE"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_download(self, make_models_url, mocked_inference_model, tmpdir):
        # arrange
        url = make_models_url(mocked_inference_model["id"])
        mock_get_response(url, mocked_inference_model)
        responses.add(
            responses.GET,
            url + "download/",
            status=200,
            content_type="application/json",
            body=b"content",
        )

        downloaded_file = tmpdir.mkdir("sub").join("download")
        downloaded_file_path = str(downloaded_file)

        # act
        model = CustomInferenceModel.get(mocked_inference_model["id"])
        model.download_latest_version(downloaded_file_path)

        # assert
        downloaded_file.read() == b"content"

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url + "download/"

    @responses.activate
    def test_assign_training_data_non_blocking(
        self, unittest_endpoint, make_models_url, mocked_inference_model, tmpdir
    ):
        status_url = "{}/status_url".format(unittest_endpoint)

        mocked_inference_model["trainingDataAssignmentInProgress"] = False
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )
        responses.add(
            responses.PATCH,
            make_models_url(mocked_inference_model["id"]) + "trainingData/",
            status=200,
            content_type="application/json",
            headers={"Location": status_url},
            body=json.dumps({"id": mocked_inference_model["id"]}),
        )
        mocked_inference_model["trainingDataAssignmentInProgress"] = True
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])

        assert not model.training_data_assignment_in_progress
        model.assign_training_data("5ea6bbc7402403321d4e1fae", max_wait=None)

        assert model.training_data_assignment_in_progress

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == make_models_url(mocked_inference_model["id"])
        assert responses.calls[1].request.method == "PATCH"
        assert (
            responses.calls[1].request.url
            == make_models_url(mocked_inference_model["id"]) + "trainingData/"
        )
        assert responses.calls[2].request.method == "GET"
        assert responses.calls[2].request.url == make_models_url(mocked_inference_model["id"])

    @responses.activate
    def test_assign_training_data_blocking(
        self, unittest_endpoint, make_models_url, mocked_inference_model, tmpdir
    ):
        status_url = "{}/status_url".format(unittest_endpoint)

        mocked_inference_model["trainingDataAssignmentInProgress"] = False
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )
        responses.add(
            responses.PATCH,
            make_models_url(mocked_inference_model["id"]) + "trainingData/",
            status=200,
            content_type="application/json",
            headers={"Location": status_url},
            body=json.dumps({"id": mocked_inference_model["id"]}),
        )
        responses.add(
            responses.GET,
            status_url,
            status=303,
            content_type="application/json",
            headers={"Location": status_url},
        )
        responses.add(
            responses.GET,
            make_models_url(mocked_inference_model["id"]),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.get(mocked_inference_model["id"])

        assert not model.training_data_assignment_in_progress
        model.assign_training_data("5ea6bbc7402403321d4e1fae")

        assert not model.training_data_assignment_in_progress

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == make_models_url(mocked_inference_model["id"])
        assert responses.calls[1].request.method == "PATCH"
        assert (
            responses.calls[1].request.url
            == make_models_url(mocked_inference_model["id"]) + "trainingData/"
        )
        assert responses.calls[2].request.method == "GET"
        assert responses.calls[2].request.url == status_url
        assert responses.calls[3].request.method == "GET"
        assert responses.calls[3].request.url == make_models_url(mocked_inference_model["id"])

    @responses.activate
    @pytest.mark.parametrize(
        "model_data",
        [
            "mocked_inference_model",
            "mocked_inference_model_no_latest_version",
            "mocked_inference_model_with_resources",
            "mocked_inference_model_with_deprecated_target_support",
        ],
    )
    def test_create(self, request, make_models_url, model_data):
        model_data = request.getfixturevalue(model_data)
        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        model = CustomInferenceModel.create(
            model_data["name"],
            TARGET_TYPE.REGRESSION,
            model_data["targetName"],
            network_egress_policy=model_data["networkEgressPolicy"],
            maximum_memory=model_data["maximumMemory"],
            replicas=model_data["replicas"],
        )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()

    def test_create_model_without_target_name(self, mocked_inference_model):
        model_data = mocked_inference_model
        with pytest.raises(
            ValueError,
            match="target_name is required for custom models with target type Regression",
        ):
            CustomInferenceModel.create(
                model_data["name"],
                TARGET_TYPE.REGRESSION,
                network_egress_policy=model_data["networkEgressPolicy"],
                maximum_memory=model_data["maximumMemory"],
                replicas=model_data["replicas"],
            )

    @responses.activate
    @pytest.mark.parametrize("labels_as_files", [True, False])
    def test_create_with_multiclass_labels(
        self, mocked_multiclass_inference_model, make_models_url, labels_as_files
    ):
        model_data = mocked_multiclass_inference_model
        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        if labels_as_files:
            with NamedTemporaryFile() as f:
                f.write("\n".join(model_data["classLabels"]).encode("utf-8"))
                f.flush()
                model = CustomInferenceModel.create(
                    model_data["name"],
                    TARGET_TYPE.MULTICLASS,
                    model_data["targetName"],
                    class_labels_file=f.name,
                    network_egress_policy=model_data["networkEgressPolicy"],
                    maximum_memory=model_data["maximumMemory"],
                    replicas=model_data["replicas"],
                )
        else:
            model = CustomInferenceModel.create(
                model_data["name"],
                TARGET_TYPE.MULTICLASS,
                model_data["targetName"],
                class_labels=model_data["classLabels"],
                network_egress_policy=model_data["networkEgressPolicy"],
                maximum_memory=model_data["maximumMemory"],
                replicas=model_data["replicas"],
            )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()
        assert (
            json.loads(responses.calls[0].request.body)["classLabels"] == model_data["classLabels"]
        )

    @responses.activate
    @pytest.mark.parametrize(
        "inference_model, target_type",
        [
            ["mocked_unstructured_inference_model", TARGET_TYPE.UNSTRUCTURED],
            ["mocked_anomaly_inference_model", TARGET_TYPE.ANOMALY],
        ],
    )
    def test_create_no_target_model(self, request, inference_model, target_type, make_models_url):
        model_data = request.getfixturevalue(inference_model)

        responses.add(
            responses.POST,
            make_models_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(model_data),
        )

        model = CustomInferenceModel.create(
            model_data["name"],
            target_type,
            network_egress_policy=model_data["networkEgressPolicy"],
            maximum_memory=model_data["maximumMemory"],
            replicas=model_data["replicas"],
        )
        assert_inference_model(model, model_data)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_models_url()

    @responses.activate
    def test_create_copy(self, make_models_url, mocked_inference_model):
        custom_model_to_copy = "5ea6bbc7402403321d4e1fad"

        url = make_models_url() + "fromCustomModel/"
        responses.add(
            responses.POST,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_inference_model),
        )

        model = CustomInferenceModel.copy_custom_model(custom_model_to_copy)
        assert_inference_model(model, mocked_inference_model)

        req = responses.calls[0].request
        assert req.method == "POST"
        assert req.url == url
        assert json.loads(req.body) == {"customModelId": custom_model_to_copy}



import trafaret as t

from ...enums import PROJECT_STAGE
from ..api_object import APIObject
from ..project import Project

__all__ = ["Image", "SampleImage", "DuplicateImage"]


class Image(APIObject):
    """An image stored in a project's dataset.

    Attributes
    ----------
    id: str
        Image ID for this image.
    image_type: str
        Image media type. Accessing this may require a server request
        and an associated delay in returning.
    image_bytes: [octet]
        Raw octets of this image. Accessing this may require a server request
        and an associated delay in returning.
    height: int
        Height of the image in pixels (72 pixels per inch).
    width: int
        Width of the image in pixels (72 pixels per inch).
    """

    _get_path = "projects/{project_id}/images/{image_id}/"
    _bytes_path = "projects/{project_id}/images/{image_id}/file/"
    _converter = t.Dict(
        {
            t.Key("image_id", optional=True): t.String(),
            t.Key("height", optional=True): t.Int(),
            t.Key("width", optional=True): t.Int(),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self.id = kwargs.get("image_id")
        self.project_id = kwargs.get("project_id")
        self.__image_type = None
        self.__image_bytes = None
        self.height = int(kwargs.get("height", 0))
        self.width = int(kwargs.get("width", 0))

    def __repr__(self):
        return (
            "datarobot.models.visualai.Image("
            "project_id={0.project_id}, "
            "image_id={0.id}, "
            "height={0.height}, "
            "width={0.width})"
        ).format(self)

    @property
    def image_type(self):
        if not self.__image_type:
            self.__get_image_bytes()
        return self.__image_type

    @property
    def image_bytes(self):
        if not self.__image_bytes:
            self.__get_image_bytes()
        return self.__image_bytes

    def __get_image_bytes(self):
        path = self._bytes_path.format(project_id=self.project_id, image_id=self.id)
        r_data = self._client.get(path)
        self.__image_type = r_data.headers.get("Content-Type")
        self.__image_bytes = r_data.content

    @classmethod
    def get(cls, project_id, image_id):
        """Get a single image object from project.

        Parameters
        ----------
        project_id: str
            Project that contains the images.
        image_id: str
            ID of image to load from the project.
        """
        path = cls._get_path.format(project_id=project_id, image_id=image_id)
        r_data = cls._client.get(path).json()
        ret = cls.from_server_data(r_data)
        ret.project_id = project_id
        return ret


class SampleImage(APIObject):
    """A sample image in a project's dataset.

    If ``Project.stage`` is ``datarobot.enums.PROJECT_STAGE.EDA2`` then
    the ``target_*`` attributes of this class will have values, otherwise
    the values will all be None.

    Attributes
    ----------
    image: Image
        Image object.
    target_value: str
        Value associated with the ``feature_name``.
    """

    _list_sample_path = "projects/{project_id}/imageSamples/"
    _list_images_path = "projects/{project_id}/images/"
    _converter = t.Dict(
        {
            t.Key("image_id", optional=True): t.String(),
            t.Key("height", optional=True): t.Int(),
            t.Key("width", optional=True): t.Int(),
            t.Key("target_value", optional=True): t.String | t.Int | t.Float | t.List(t.String),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self.image = Image(**kwargs)
        self.project_id = kwargs.get("project_id")
        self.target_value = kwargs.get("target_value")

    def __repr__(self):
        return (
            "datarobot.models.visualai.SampleImage("
            "project_id={0.project_id}, "
            "image_id={0.image.id}, "
            "target_value={0.target_value})"
        ).format(self)

    @classmethod
    def list(cls, project_id, feature_name, target_value=None, offset=None, limit=None):
        """Get sample images from a project.

        Parameters
        ----------
        project_id: str
            Project that contains the images.
        feature_name: str
            Name of feature column that contains images.
        target_value: str
            Target value to filter images.
        offset: int
            Number of images to be skipped.
        limit: int
            Number of images to be returned.
        """
        project = Project.get(project_id)
        list_params = {}
        if project.stage in [PROJECT_STAGE.EDA2, PROJECT_STAGE.MODELING]:
            path = cls._list_images_path.format(project_id=project_id)
        else:
            path = cls._list_sample_path.format(project_id=project_id)
            list_params["featureName"] = feature_name
        if target_value:
            list_params["targetValue"] = target_value
        if offset:
            list_params["offset"] = int(offset)
        if limit:
            list_params["limit"] = int(limit)
        r_data = cls._client.get(path, params=list_params).json()
        ret = []
        for si_data in r_data["data"]:
            si = cls.from_server_data(si_data)
            si.project_id = project_id
            si.image.project_id = project_id
            ret.append(si)
        return ret


class DuplicateImage(APIObject):
    """An image that was duplicated in the project dataset.

    Attributes
    ----------
    image: Image
        Image object.
    count: int
        Number of times the image was duplicated.
    """

    _list_path = "projects/{project_id}/duplicateImages/{feature_name}/"
    _converter = t.Dict(
        {t.Key("image_id", optional=True): t.String(), t.Key("row_count", optional=True): t.Int()}
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self.image = Image(**kwargs)
        self.project_id = kwargs.get("project_id")
        self.count = kwargs.get("row_count")

    def __repr__(self):
        return (
            "datarobot.models.visualai.DuplicateImage("
            "project_id={0.project_id}, "
            "image_id={0.image.id}, "
            "count={0.count})"
        ).format(self)

    @classmethod
    def list(cls, project_id, feature_name, offset=None, limit=None):
        """Get all duplicate images in a project.

        Parameters
        ----------
        project_id: str
            Project that contains the images.
        feature_name: str
            Name of feature column that contains images.
        offset: int
            Number of images to be skipped.
        limit: int
            Number of images to be returned.
        """
        path = cls._list_path.format(project_id=project_id, feature_name=feature_name)
        list_params = {}
        if offset:
            list_params["offset"] = int(offset)
        if limit:
            list_params["limit"] = int(limit)
        r_data = cls._client.get(path, params=list_params).json()
        ret = []
        for si_data in r_data["data"]:
            si = cls.from_server_data(si_data)
            si.project_id = project_id
            si.image.project_id = project_id
            ret.append(si)
        return ret



import json

import responses

from datarobot import BlenderModel, ModelJob, Project
from datarobot.enums import BLENDER_METHOD


@responses.activate
def test_blend_return_job_id():
    job_id = 1
    location = "https://host_name.com/projects/p-id/modelJobs/{}/".format(job_id)
    model1 = "57e27abc8e43553e572f8df7"
    model2 = "57e27abc8e43553e572f8df8"
    responses.add(
        responses.POST,
        "https://host_name.com/projects/p-id/blenderModels/",
        status=202,
        body="",
        adding_headers={"Location": location},
    )
    model_job_url = "https://host_name.com/projects/p-id/modelJobs/{}/".format(job_id)
    model_job_data = {
        "status": "inprogress",
        "processes": [],
        "projectId": "p-id",
        "samplePct": 28.349,
        "modelType": "Naive Bayes combiner classifier",
        "featurelistId": "56d8620bccf94e26f37af0a3",
        "modelCategory": "model",
        "blueprintId": "2a1b9ae97fe61880332e196c770c1f9f",
        "isBlocked": False,
        "id": job_id,
    }
    responses.add(
        responses.GET,
        model_job_url,
        status=200,
        body=json.dumps(model_job_data),
        content_type="application/json",
    )
    result = Project("p-id").blend([model1, model2], "AVG")
    assert isinstance(result, ModelJob)
    assert result.id == job_id


@responses.activate
def test_check_blend(project, project_url):
    model1 = "1234deadbeefdeadbeef4321"
    model2 = "feeddeadcafebeef43211234"
    url = "{}blenderModels/blendCheck/".format(project_url)
    response_body = {"blendable": False, "reason": "because I say so"}
    responses.add(responses.POST, url, json=response_body, content_type="application/json")
    result = project.check_blendable([model1, model2], BLENDER_METHOD.AVERAGE)
    assert result.supported is False
    assert result.reason == response_body["reason"]
    assert all(
        substring in result.context
        for substring in ["blendability", model1, model2, BLENDER_METHOD.AVERAGE]
    )


@responses.activate
def test_get_blenders_return_valid_objects(blenders_list_response_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/blenderModels/",
        body=blenders_list_response_json,
        status=200,
        content_type="application/json",
    )
    blenders = Project("p-id").get_blenders()
    assert all(isinstance(blender, BlenderModel) for blender in blenders)



from collections import namedtuple
import os

import dateutil
from pandas import DataFrame  # noqa F401
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.credential import CredentialDataSchema
from datarobot.models.feature import DatasetFeature
from datarobot.models.featurelist import DatasetFeaturelist
from datarobot.models.project import Project
from datarobot.utils import dataframe_to_buffer, encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate
from datarobot.utils.sourcedata import list_of_records_to_buffer
from datarobot.utils.waiters import wait_for_async_resolution

from ..enums import DEFAULT_MAX_WAIT, DEFAULT_TIMEOUT

ProjectLocation = namedtuple("ProjectLocation", ["url", "id"])

FeatureTypeCount = namedtuple("FeatureTypeCount", ["count", "feature_type"])


_base_dataset_schema = t.Dict(
    {
        t.Key("dataset_id"): t.String,
        t.Key("version_id"): t.String,
        t.Key("name"): t.String,
        t.Key("categories"): t.List(t.String),
        t.Key("creation_date") >> "created_at": t.Call(lambda s: dateutil.parser.parse(s)),
        t.Key("created_by"): t.String,
        t.Key("data_persisted", optional=True): t.Bool,
        t.Key("is_data_engine_eligible"): t.Bool,
        t.Key("is_latest_version"): t.Bool,
        t.Key("is_snapshot"): t.Bool,
        t.Key("dataset_size", optional=True) >> "size": t.Int,
        t.Key("row_count", optional=True): t.Int,
        t.Key("processing_state"): t.String,
    }
)


class Dataset(APIObject):
    """ Represents a Dataset returned from the api/v2/datasets/ endpoints.

    Attributes
    ----------
    id: string
        The ID of this dataset
    name: string
        The name of this dataset in the catalog
    is_latest_version: bool
        Whether this dataset version is the latest version
        of this dataset
    version_id: string
        The object ID of the catalog_version the dataset belongs to
    categories: list(string)
        An array of strings describing the intended use of the dataset. The
        supported options are "TRAINING" and "PREDICTION".
    created_at: string
        The date when the dataset was created
    created_by: string
        Username of the user who created the dataset
    is_snapshot: bool
        Whether the dataset version is an immutable snapshot of data
        which has previously been retrieved and saved to Data_robot
    data_persisted: bool, optional
        If true, user is allowed to view extended data profile
        (which includes data statistics like min/max/median/mean, histogram, etc.) and download
        data. If false, download is not allowed and only the data schema (feature names and types)
        will be available.
    is_data_engine_eligible: bool
        Whether this dataset can be
        a data source of a data engine query.
    processing_state: string
        Current ingestion process state of
        the dataset
    row_count: int, optional
        The number of rows in the dataset.
    size: int, optional
        The size of the dataset as a CSV in bytes.
    """

    _converter = _base_dataset_schema.allow_extra("*")
    _path = "datasets/"

    def __init__(
        self,
        dataset_id,
        version_id,
        name,
        categories,
        created_at,
        created_by,
        is_data_engine_eligible,
        is_latest_version,
        is_snapshot,
        processing_state,
        data_persisted=None,
        size=None,
        row_count=None,
    ):
        self.id = dataset_id
        self.version_id = version_id
        self.name = name
        self.data_persisted = data_persisted
        self.categories = categories[:]
        self.created_at = created_at
        self.created_by = created_by
        self.is_data_engine_eligible = is_data_engine_eligible
        self.is_latest_version = is_latest_version
        self.is_snapshot = is_snapshot
        self.size = size
        self.row_count = row_count
        self.processing_state = processing_state

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}(name={!r}, id={!r})".format(self.__class__.__name__, self.name, self.id)
        )

    @classmethod
    def create_from_file(
        cls,
        file_path=None,
        filelike=None,
        categories=None,
        read_timeout=DEFAULT_TIMEOUT.UPLOAD,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset from a file. Returns when the dataset has
        been successfully uploaded and processed.

        Warning: This function does not clean up it's open files. If you pass a filelike, you are
        responsible for closing it. If you pass a file_path, this will create a file object from
        the file_path but will not close it.

        Parameters
        ----------
        file_path: string, optional
            The path to the file. This will create a file object pointing to that file but will
            not close it.
        filelike: file, optional
            An open and readable file object.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: int, optional
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            A fully armed and operational Dataset
        """
        _assert_single_parameter(("filelike", "file_path"), file_path, filelike)

        upload_url = "{}fromFile/".format(cls._path)
        default_fname = "data.csv"
        if file_path:
            fname = os.path.basename(file_path)
            response = cls._client.build_request_with_file(
                fname=fname,
                file_path=file_path,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )
        else:
            try:
                fname = filelike.name
            except AttributeError:
                fname = default_fname
            response = cls._client.build_request_with_file(
                fname=fname,
                filelike=filelike,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        dataset = cls.from_location(new_dataset_location)
        if categories:
            dataset.modify(categories=categories)
        return dataset

    @classmethod
    def create_from_in_memory_data(
        cls,
        data_frame=None,
        records=None,
        categories=None,
        read_timeout=DEFAULT_TIMEOUT.UPLOAD,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset from in-memory data. Returns when the dataset has
        been successfully uploaded and processed.

        The data can be either a pandas DataFrame or a list of dictionaries with identical keys.

        Parameters
        ----------
        data_frame: DataFrame, optional
            The data frame to upload
        records: list[dict], optional
            A list of dictionaries with identical keys to upload
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: int, optional
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        _assert_single_parameter(("data_frame", "records"), data_frame, records)
        if data_frame is not None:
            buff = dataframe_to_buffer(data_frame)
        else:
            buff = list_of_records_to_buffer(records)
        return cls.create_from_file(
            filelike=buff, categories=categories, read_timeout=read_timeout, max_wait=max_wait,
        )

    @classmethod
    def create_from_url(
        cls,
        url,
        do_snapshot=None,
        persist_data_after_ingestion=None,
        categories=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset from data stored at a url.
        Returns when the dataset has been successfully uploaded and processed.

        Parameters
        ----------
        url: string
            The URL to use as the source of data for the dataset being created.
        do_snapshot: bool, optional
            If unset, uses the server default: True.
            If true, creates a snapshot dataset; if
            false, creates a remote dataset. Creating snapshots from non-file sources requires an
            additional permission, `Enable Create Snapshot Data Source`.
        persist_data_after_ingestion: bool, optional
            If unset, uses the server default: True.
            If true, will enforce saving all data
            (for download and sampling) and will allow a user to view extended data profile
            (which includes data statistics like min/max/median/mean, histogram, etc.). If false,
            will not enforce saving data. The data schema (feature names and types) still will be
            available. Specifying this parameter to false and `doSnapshot` to true will result in
            an error.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        base_data = {
            "url": url,
            "do_snapshot": do_snapshot,
            "persist_data_after_ingestion": persist_data_after_ingestion,
            "categories": categories,
        }
        data = _remove_empty_params(base_data)
        upload_url = "{}fromURL/".format(cls._path)
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    def create_from_data_source(
        cls,
        data_source_id,
        username=None,
        password=None,
        do_snapshot=None,
        persist_data_after_ingestion=None,
        categories=None,
        credential_id=None,
        use_kerberos=None,
        credential_data=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset from data stored at a DataSource.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: v2.22

        Parameters
        ----------
        data_source_id: string
            The ID of the DataSource to use as the source of data.
        username: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored.
        do_snapshot: bool, optional
            If unset, uses the server default: True.
            If true, creates a snapshot dataset; if
            false, creates a remote dataset. Creating snapshots from non-file sources requires an
            additional permission, `Enable Create Snapshot Data Source`.
        persist_data_after_ingestion: bool, optional
            If unset, uses the server default: True.
            If true, will enforce saving all data
            (for download and sampling) and will allow a user to view extended data profile
            (which includes data statistics like min/max/median/mean, histogram, etc.). If false,
            will not enforce saving data. The data schema (feature names and types) still will be
            available. Specifying this parameter to false and `doSnapshot` to true will result in
            an error.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        credential_id: string, optional
            The ID of the set of credentials to
            use instead of user and password. Note that with this change, username and password
            will become optional.
        use_kerberos: bool, optional
            If unset, uses the server default: False.
            If true, use kerberos authentication for database authentication.
        credential_data: dict, optional
            The credentials to authenticate with the database, to use instead of user/password or
            credential ID.
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful


        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        base_data = {
            "data_source_id": data_source_id,
            "user": username,
            "password": password,
            "do_snapshot": do_snapshot,
            "persist_data_after_ingestion": persist_data_after_ingestion,
            "categories": categories,
            "credential_id": credential_id,
            "use_kerberos": use_kerberos,
            "credential_data": credential_data,
        }
        data = _remove_empty_params(base_data)

        if "credential_data" in data:
            data["credential_data"] = CredentialDataSchema(data["credential_data"])

        upload_url = "{}fromDataSource/".format(cls._path)
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    def get(cls, dataset_id):
        """Get information about a dataset.

        Parameters
        ----------
        dataset_id : string
            the id of the dataset

        Returns
        -------
        dataset : Dataset
            the queried dataset
        """

        path = "{}{}/".format(cls._path, dataset_id)
        return cls.from_location(path)

    @classmethod
    def delete(cls, dataset_id):
        """
        Soft deletes a dataset.  You cannot get it or list it or do actions with it, except for
        un-deleting it.

        Parameters
        ----------
        dataset_id: string
            The id of the dataset to mark for deletion

        Returns
        -------
        None
        """
        path = "{}{}/".format(cls._path, dataset_id)
        cls._client.delete(path)

    @classmethod
    def un_delete(cls, dataset_id):
        """
        Un-deletes a previously deleted dataset.  If the dataset was not deleted, nothing happens.

        Parameters
        ----------
        dataset_id: string
            The id of the dataset to un-delete

        Returns
        -------
        None
        """
        path = "{}{}/deleted/".format(cls._path, dataset_id)
        cls._client.patch(path)

    @classmethod
    def list(cls, category=None, filter_failed=None, order_by=None):
        """
        List all datasets a user can view.


        Parameters
        ----------
        category: string, optional
            Optional. If specified, only dataset versions that have the specified category will be
            included in the results. Categories identify the intended use of the dataset; supported
            categories are "TRAINING" and "PREDICTION".

        filter_failed: bool, optional
            If unset, uses the server default: False.
            Whether datasets that failed during import should be excluded from the results.
            If True invalid datasets will be excluded.

        order_by: string, optional
            If unset, uses the server default: "-created".
            Sorting order which will be applied to catalog list, valid options are:
            - "created" -- ascending order by creation datetime;
            - "-created" -- descending order by creation datetime.

        Returns
        -------
        list[Dataset]
            a list of datasets the user can view

        """
        return list(cls.iterate(category=category, order_by=order_by, filter_failed=filter_failed))

    @classmethod
    def iterate(cls, offset=None, limit=None, category=None, order_by=None, filter_failed=None):
        """
        Get an iterator for the requested datasets a user can view.
        This lazily retrieves results. It does not get the next page from the server until the
        current page is exhausted.

        Parameters
        ----------
        offset: int, optional
            If set, this many results will be skipped

        limit: int, optional
            Specifies the size of each page retrieved from the server.  If unset, uses the server
            default.

        category: string, optional
            Optional. If specified, only dataset versions that have the specified category will be
            included in the results. Categories identify the intended use of the dataset; supported
            categories are "TRAINING" and "PREDICTION".

        filter_failed: bool, optional
            If unset, uses the server default: False.
            Whether datasets that failed during import should be excluded from the results.
            If True invalid datasets will be excluded.

        order_by: string, optional
            If unset, uses the server default: "-created".
            Sorting order which will be applied to catalog list, valid options are:
            - "created" -- ascending order by creation datetime;
            - "-created" -- descending order by creation datetime.

        Yields
        -------
        Dataset
            An iterator of the datasets the user can view

        """
        all_params = {
            "offset": offset,
            "limit": limit,
            "category": category,
            "order_by": order_by,
            "filter_failed": filter_failed,
        }
        params = _remove_empty_params(all_params)
        _update_filter_failed(params)

        for dataset_json in unpaginate(cls._path, params, cls._client):
            yield cls.from_server_data(dataset_json)

    def update(self):
        """
        Updates the Dataset attributes in place with the latest information from the server.

        Returns
        -------
        None
        """
        new_dataset = self.get(self.id)
        update_attrs = (
            "name",
            "created_by",
            "created_at",
            "version_id",
            "is_latest_version",
            "is_snapshot",
            "data_persisted",
            "categories",
            "size",
            "row_count",
            "processing_state",
        )
        for attr in update_attrs:
            setattr(self, attr, getattr(new_dataset, attr))

    def modify(self, name=None, categories=None):
        """
        Modifies the Dataset name and/or categories.  Updates the object in place.

        Parameters
        ----------
        name: string, optional
            The new name of the dataset

        categories: list[string], optional
            A list of strings describing the intended use of the
            dataset. The supported options are "TRAINING" and "PREDICTION". If any
            categories were previously specified for the dataset, they will be overwritten.

        Returns
        -------
        None

        """
        if name is None and categories is None:
            return

        url = "{}{}/".format(self._path, self.id)
        params = {"name": name, "categories": categories}
        params = _remove_empty_params(params)

        response = self._client.patch(url, data=params)
        data = response.json()
        self.name = data["name"]
        self.categories = data["categories"]

    def get_details(self):
        """
        Gets the details for this Dataset

        Returns
        -------
        DatasetDetails
        """
        return DatasetDetails.get(self.id)

    def get_all_features(self, order_by=None):
        """
        Get a list of all the features for this dataset.

        Parameters
        ----------
        order_by: string, optional
            If unset, uses the server default: 'name'.
            How the features should be ordered. Can be 'name' or 'featureType'.

        Returns
        -------
        list[DatasetFeature]
        """
        return list(self.iterate_all_features(order_by=order_by))

    def iterate_all_features(self, offset=None, limit=None, order_by=None):
        """
        Get an iterator for the requested features of a dataset.
        This lazily retrieves results. It does not get the next page from the server until the
        current page is exhausted.

        Parameters
        ----------
        offset: int, optional
            If set, this many results will be skipped.

        limit: int, optional
            Specifies the size of each page retrieved from the server.  If unset, uses the server
            default.

        order_by: string, optional
            If unset, uses the server default: 'name'.
            How the features should be ordered. Can be 'name' or 'featureType'.

        Yields
        -------
        DatasetFeature
        """
        all_params = {
            "offset": offset,
            "limit": limit,
            "order_by": order_by,
        }
        params = _remove_empty_params(all_params)

        url = "{}{}/allFeaturesDetails/".format(self._path, self.id)
        for dataset_json in unpaginate(url, params, self._client):
            yield DatasetFeature.from_server_data(dataset_json)

    def get_featurelists(self):
        """
        Get DatasetFeaturelists created on this Dataset

        Returns
        -------
        feature_lists: list[DatasetFeaturelist]
        """
        url = "{}{}/featurelists/".format(self._path, self.id)
        params = {}
        result = unpaginate(url, params, self._client)
        return [DatasetFeaturelist.from_server_data(el) for el in result]

    def create_featurelist(self, name, features):
        """ Create a new dataset featurelist

        Parameters
        ----------
        name : str
            the name of the modeling featurelist to create. Names must be unique within the
            dataset, or the server will return an error.
        features : list of str
            the names of the features to include in the dataset featurelist. Each feature must
            be a dataset feature.

        Returns
        -------
        featurelist : DatasetFeaturelist
            the newly created featurelist

        Examples
        --------
        .. code-block:: python

            dataset = Dataset.get('1234deadbeeffeeddead4321')
            dataset_features = dataset.get_all_features()
            selected_features = [feat.name for feat in dataset_features][:5]  # select first five
            new_flist = dataset.create_featurelist('Simple Features', selected_features)
        """
        url = "{}{}/featurelists/".format(self._path, self.id)

        payload = {"name": name, "features": features}
        response = self._client.post(url, data=payload)
        return DatasetFeaturelist.from_server_data(response.json())

    def get_file(self, file_path=None, filelike=None):
        """
        Retrieves all the originally uploaded data in CSV form.
        Writes it to either the file or a filelike object that can write bytes.

        Only one of file_path or filelike can be provided and it must be provided as a
        keyword argument (i.e. file_path='path-to-write-to'). If a file-like object is
        provided, the user is responsible for closing it when they are done.

        The user must also have permission to download data.

        Parameters
        ----------
        file_path: string, optional
            The destination to write the file to.
        filelike: file, optional
            A file-like object to write to.  The object must be able to write bytes. The user is
            responsible for closing the object

        Returns
        -------
        None
        """
        _assert_single_parameter(("filelike", "file_path"), filelike, file_path)

        response = self._client.get("{}{}/file/".format(self._path, self.id))
        if file_path:
            with open(file_path, "wb") as f:
                f.write(response.content)
        if filelike:
            filelike.write(response.content)

    def get_projects(self):
        """
        Retrieves the Dataset's projects as ProjectLocation named tuples.

        Returns
        -------
        locations: list[ProjectLocation]
        """
        url = "{}{}/projects/".format(self._path, self.id)
        return [ProjectLocation(**kwargs) for kwargs in unpaginate(url, None, self._client)]

    def create_project(
        self,
        project_name=None,
        user=None,
        password=None,
        credential_id=None,
        use_kerberos=None,
        credential_data=None,
    ):
        """
        Create a :class:`datarobot.models.Project` from this dataset

        Parameters
        ----------
        project_name: string, optional
            The name of the project to be created.
            If not specified, will be "Untitled Project" for database connections, otherwise
            the project name will be based on the file used.
        user: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored
        credential_id: string, optional
            The ID of the set of credentials to use instead of user and password.
        use_kerberos: bool, optional
            Server default is False.
            If true, use kerberos authentication for database authentication.
        credential_data: dict, optional
            The credentials to authenticate with the database, to use instead of user/password or
            credential ID.

        Returns
        -------
        Project
        """
        return Project.create_from_dataset(
            self.id,
            dataset_version_id=self.version_id,
            project_name=project_name,
            user=user,
            password=password,
            credential_id=credential_id,
            use_kerberos=use_kerberos,
            credential_data=credential_data,
        )

    @classmethod
    def create_version_from_file(
        cls,
        dataset_id,
        file_path=None,
        filelike=None,
        categories=None,
        read_timeout=DEFAULT_TIMEOUT.UPLOAD,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset version from a file. Returns when the new dataset
        version has been successfully uploaded and processed.

        Warning: This function does not clean up it's open files. If you pass a filelike, you are
        responsible for closing it. If you pass a file_path, this will create a file object from
        the file_path but will not close it.

        .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        file_path: string, optional
            The path to the file. This will create a file object pointing to that file but will
            not close it.
        filelike: file, optional
            An open and readable file object.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: int, optional
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            A fully armed and operational Dataset version
        """
        _assert_single_parameter(("filelike", "file_path"), file_path, filelike)

        upload_url = "{}{}/versions/fromFile/".format(cls._path, dataset_id)
        default_fname = "data.csv"
        if file_path:
            fname = os.path.basename(file_path)
            response = cls._client.build_request_with_file(
                fname=fname,
                file_path=file_path,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )
        else:
            try:
                fname = filelike.name
            except AttributeError:
                fname = default_fname
            response = cls._client.build_request_with_file(
                fname=fname,
                filelike=filelike,
                url=upload_url,
                read_timeout=read_timeout,
                method="post",
            )

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        dataset = cls.from_location(new_dataset_location)
        if categories:
            dataset.modify(categories=categories)
        return dataset

    @classmethod
    def create_version_from_in_memory_data(
        cls,
        dataset_id,
        data_frame=None,
        records=None,
        categories=None,
        read_timeout=DEFAULT_TIMEOUT.UPLOAD,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset version for a dataset from in-memory data.
        Returns when the dataset has been successfully uploaded and processed.

        The data can be either a pandas DataFrame or a list of dictionaries with identical keys.

         .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        data_frame: DataFrame, optional
            The data frame to upload
        records: list[dict], optional
            A list of dictionaries with identical keys to upload
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        read_timeout: int, optional
            The maximum number of seconds to wait for the server to respond indicating that the
            initial upload is complete
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        _assert_single_parameter(("data_frame", "records"), data_frame, records)
        if data_frame is not None:
            buff = dataframe_to_buffer(data_frame)
        else:
            buff = list_of_records_to_buffer(records)
        return cls.create_version_from_file(
            dataset_id,
            filelike=buff,
            categories=categories,
            read_timeout=read_timeout,
            max_wait=max_wait,
        )

    @classmethod
    def create_version_from_url(cls, dataset_id, url, categories=None, max_wait=DEFAULT_MAX_WAIT):
        """
        A blocking call that creates a new Dataset from data stored at a url for a given dataset.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        url: string
            The URL to use as the source of data for the dataset being created.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        base_data = {
            "url": url,
            "categories": categories,
        }
        data = _remove_empty_params(base_data)
        upload_url = "{}{}/versions/fromURL/".format(cls._path, dataset_id)
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)

    @classmethod
    def create_version_from_data_source(
        cls,
        dataset_id,
        data_source_id,
        username=None,
        password=None,
        categories=None,
        credential_id=None,
        use_kerberos=None,
        credential_data=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """
        A blocking call that creates a new Dataset from data stored at a DataSource.
        Returns when the dataset has been successfully uploaded and processed.

        .. versionadded:: v2.23

        Parameters
        ----------
        dataset_id: string
            The ID of the dataset for which new version to be created
        data_source_id: string
            The ID of the DataSource to use as the source of data.
        username: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        credential_id: string, optional
            The ID of the set of credentials to
            use instead of user and password. Note that with this change, username and password
            will become optional.
        use_kerberos: bool, optional
            If unset, uses the server default: False.
            If true, use kerberos authentication for database authentication.
        credential_data: dict, optional
            The credentials to authenticate with the database, to use instead of user/password or
            credential ID.
        max_wait: int, optional
            Time in seconds after which project creation is considered unsuccessful

        Returns
        -------
        response: Dataset
            The Dataset version created from the uploaded data
        """
        base_data = {
            "data_source_id": data_source_id,
            "user": username,
            "password": password,
            "categories": categories,
            "credential_id": credential_id,
            "use_kerberos": use_kerberos,
            "credential_data": credential_data,
        }
        data = _remove_empty_params(base_data)

        if "credential_data" in data:
            data["credential_data"] = CredentialDataSchema(data["credential_data"])

        upload_url = "{}{}/versions/fromDataSource/".format(cls._path, dataset_id)
        response = cls._client.post(upload_url, data=data)

        new_dataset_location = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.from_location(new_dataset_location)


def _assert_single_parameter(param_names, *params):
    if sum(param is not None for param in params) != 1:
        raise TypeError("One and only parameter of: {}".format(param_names))


def _remove_empty_params(params_dict):
    return {key: value for key, value in params_dict.items() if value is not None}


def _update_filter_failed(query_params):
    try:
        key = "filter_failed"
        query_params[key] = str(query_params[key]).lower()
    except KeyError:
        pass


def _safe_merge(first, second):
    # type: (t.Dict, t.Dict) -> t.Dict

    second_names = {el.name for el in second.keys}
    if any(el.name in second_names for el in first.keys):
        raise ValueError("Duplicate keys detected")

    return first.merge(second)


class DatasetDetails(APIObject):
    """ Represents a detailed view of a Dataset. The `to_dataset` method creates a Dataset
    from this details view.

    Attributes
    ----------
    dataset_id: string
        The ID of this dataset
    name: string
        The name of this dataset in the catalog
    is_latest_version: bool
        Whether this dataset version is the latest version
        of this dataset
    version_id: string
        The object ID of the catalog_version the dataset belongs to
    categories: list(string)
        An array of strings describing the intended use of the dataset. The
        supported options are "TRAINING" and "PREDICTION".
    created_at: string
        The date when the dataset was created
    created_by: string
        Username of the user who created the dataset
    is_snapshot: bool
        Whether the dataset version is an immutable snapshot of data
        which has previously been retrieved and saved to Data_robot
    data_persisted: bool, optional
        If true, user is allowed to view extended data profile
        (which includes data statistics like min/max/median/mean, histogram, etc.) and download
        data. If false, download is not allowed and only the data schema (feature names and types)
        will be available.
    is_data_engine_eligible: bool
        Whether this dataset can be
        a data source of a data engine query.
    processing_state: string
        Current ingestion process state of
        the dataset
    row_count: int, optional
        The number of rows in the dataset.
    size: int, optional
        The size of the dataset as a CSV in bytes.
    data_engine_query_id: string, optional
        ID of the source data engine query
    data_source_id: string, optional
        ID of the datasource used as the source of the dataset
    data_source_type: string
        the type of the datasource that was used as the source of the
        dataset
    description: string, optional
        the description of the dataset
    eda1_modification_date: string, optional
        the ISO 8601 formatted date and time when the EDA1 for
        the dataset was updated
    eda1_modifier_full_name: string, optional
        the user who was the last to update EDA1 for the
        dataset
    error: string
        details of exception raised during ingestion process, if any
    feature_count: int, optional
        total number of features in the dataset
    feature_count_by_type: list[FeatureTypeCount]
        number of features in the dataset grouped by feature type
    last_modification_date: string
        the ISO 8601 formatted date and time when the dataset
        was last modified
    last_modifier_full_name: string
        full name of user who was the last to modify the
        dataset
    tags: list[string]
        list of tags attached to the item
    uri: string
        the uri to datasource like:
        - 'file_name.csv'
        - 'jdbc:DATA_SOURCE_GIVEN_NAME/SCHEMA.TABLE_NAME'
        - 'jdbc:DATA_SOURCE_GIVEN_NAME/<query>' - for `query` based datasources
        - 'https://s3.amazonaws.com/datarobot_test/kickcars-sample-200.csv'
        - etc.
    """

    _extra_fields = t.Dict(
        {
            t.Key("data_engine_query_id", optional=True): t.String,
            t.Key("data_source_id", optional=True): t.String,
            t.Key("data_source_type"): t.String(allow_blank=True),
            t.Key("description", optional=True): t.String,
            t.Key("eda1_modification_date", optional=True): t.Call(
                lambda s: dateutil.parser.parse(s)
            ),
            t.Key("eda1_modifier_full_name", optional=True): t.String,
            t.Key("error"): t.String(allow_blank=True),
            t.Key("feature_count", optional=True): t.Int,
            t.Key("feature_count_by_type", optional=True): t.List(
                t.Call(lambda d: FeatureTypeCount(**d))
            ),
            t.Key("last_modification_date"): t.Call(lambda s: dateutil.parser.parse(s)),
            t.Key("last_modifier_full_name"): t.String,
            t.Key("tags", optional=True): t.List(t.String),
            t.Key("uri"): t.String,
        }
    )

    _converter = _safe_merge(_extra_fields, _base_dataset_schema).allow_extra("*")

    _path = "datasets/"

    def __init__(
        self,
        dataset_id,
        version_id,
        categories,
        created_by,
        created_at,
        data_source_type,
        error,
        is_latest_version,
        is_snapshot,
        is_data_engine_eligible,
        last_modification_date,
        last_modifier_full_name,
        name,
        uri,
        data_persisted=None,
        data_engine_query_id=None,
        data_source_id=None,
        description=None,
        eda1_modification_date=None,
        eda1_modifier_full_name=None,
        feature_count=None,
        feature_count_by_type=None,
        processing_state=None,
        row_count=None,
        size=None,
        tags=None,
    ):
        self.dataset_id = dataset_id
        self.version_id = version_id
        self.categories = categories
        self.created_by = created_by
        self.created_at = created_at
        self.data_source_type = data_source_type
        self.error = error
        self.is_latest_version = is_latest_version
        self.is_snapshot = is_snapshot
        self.is_data_engine_eligible = is_data_engine_eligible
        self.last_modification_date = last_modification_date
        self.last_modifier_full_name = last_modifier_full_name
        self.name = name
        self.uri = uri
        self.data_persisted = data_persisted
        self.data_engine_query_id = data_engine_query_id
        self.data_source_id = data_source_id
        self.description = description
        self.eda1_modification_date = eda1_modification_date
        self.eda1_modifier_full_name = eda1_modifier_full_name
        self.feature_count = feature_count
        self.feature_count_by_type = feature_count_by_type
        self.processing_state = processing_state
        self.row_count = row_count
        self.size = size
        self.tags = tags

    @classmethod
    def get(cls, dataset_id):
        """
        Get details for a Dataset from the server

        Parameters
        ----------
        dataset_id: str
            The id for the Dataset from which to get details

        Returns
        -------
        DatasetDetails
        """
        path = "{}{}/".format(cls._path, dataset_id)
        return cls.from_location(path)

    def to_dataset(self):
        """
        Build a Dataset object from the information in this object

        Returns
        -------
        Dataset
        """
        return Dataset(
            dataset_id=self.dataset_id,
            name=self.name,
            created_at=self.created_at,
            created_by=self.created_by,
            version_id=self.version_id,
            categories=self.categories,
            is_latest_version=self.is_latest_version,
            is_data_engine_eligible=self.is_data_engine_eligible,
            is_snapshot=self.is_snapshot,
            data_persisted=self.data_persisted,
            size=self.size,
            row_count=self.row_count,
            processing_state=self.processing_state,
        )



import json

import pytest
import responses

from datarobot import errors
from datarobot.utils.waiters import wait_for_async_resolution, wait_for_custom_resolution


class TestWaitForAsyncResolution(object):
    @responses.activate
    def test_success(self, client, async_url, async_success_callback, resolved_url):
        responses.add_callback(
            responses.GET,
            async_url,
            callback=async_success_callback,
            content_type="application/json",
        )
        result = wait_for_async_resolution(client, async_url)
        assert result == resolved_url

    @responses.activate
    def test_job_failure(
        self, client, async_url, async_job_failure_callback,
    ):
        responses.add_callback(
            responses.GET,
            async_url,
            callback=async_job_failure_callback,
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncProcessUnsuccessfulError):
            wait_for_async_resolution(client, async_url)

    @responses.activate
    def test_server_wtf(
        self, client, async_url, async_server_nonsense_callback,
    ):
        responses.add_callback(
            responses.GET,
            async_url,
            callback=async_server_nonsense_callback,
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncFailureError):
            wait_for_async_resolution(client, async_url)

    @responses.activate
    def test_server_timeout(self, client, async_url, async_not_finished_callback, mock_async_time):
        responses.add_callback(
            responses.GET,
            async_url,
            callback=async_not_finished_callback,
            content_type="application/json",
        )
        mock_async_time.time.side_effect = (0, 0, 10 ** 6)

        with pytest.raises(errors.AsyncTimeoutError):
            wait_for_async_resolution(client, async_url)


class TestWaitForCustomResolution(object):
    @responses.activate
    @pytest.mark.parametrize(
        "callback", ["async_job_failure_callback", "async_job_completed_callback"],
    )
    def test_success(self, request, client, async_url, callback, custom_resolver):
        callback = request.getfixturevalue(callback)
        responses.add_callback(
            responses.GET, async_url, callback=callback, content_type="application/json"
        )
        result = wait_for_custom_resolution(client, async_url, custom_resolver)
        _, _, expected_body = callback(None)
        assert result == json.loads(expected_body)

    @responses.activate
    def test_server_timeout(
        self, client, async_url, async_not_finished_callback, mock_async_time, custom_resolver
    ):
        responses.add_callback(
            responses.GET,
            async_url,
            callback=async_not_finished_callback,
            content_type="application/json",
        )
        mock_async_time.time.side_effect = (0, 0, 10 ** 6)

        with pytest.raises(errors.AsyncTimeoutError):
            wait_for_custom_resolution(client, async_url, custom_resolver)



import os
import unittest

import six.moves.urllib.parse as urllibparse

from datarobot.helpers.eligibility_result import EligibilityResult
from datarobot.helpers.partitioning_methods import (
    BasePartitioningMethod,
    DatetimePartitioningSpecification,
    GroupTVH,
)

_here = os.path.dirname(os.path.abspath(__file__))
FIXTURES_DIR = os.path.join(_here, "fixtures")


def fixture_file_path(filename):
    return os.path.join(FIXTURES_DIR, filename)


class URLParamsTestCase(unittest.TestCase):
    def assert_has_params(self, url, **params):
        query_string = urllibparse.unquote(url).split("?")[1]
        query_parts = query_string.split("&")
        query_pairs = [param.split("=") for param in query_parts]
        query_params = {pair[0]: pair[1] for pair in query_pairs}

        for key, value in query_params.items():
            self.assertIn(key, query_params, "Expected parameter {} not found".format(key))
            self.assertEqual(query_params[key], params[key])


class HelpersTestCase(unittest.TestCase):
    def test_cant_create_without_cv(self):
        with self.assertRaises(TypeError):
            BasePartitioningMethod(validation_type="CV")

        with self.assertRaises(TypeError):
            BasePartitioningMethod(cv_method="user")

    def test_partitioning_repr(self):
        p_method = GroupTVH(
            holdout_pct=1,
            validation_pct=2,
            partition_key_cols=["one_feature", "sec feature"],
            seed=42,
        )
        assert repr(p_method).startswith("GroupTVH({")

    def test_create_with_right(self):
        p_method = BasePartitioningMethod(validation_type="CV", cv_method="user")
        self.assertEqual(p_method.validation_type, "CV")
        self.assertEqual(p_method.cv_method, "user")

    def test_collect_payload(self):
        p_method = BasePartitioningMethod(validation_type="CV", cv_method="user")
        p = p_method.collect_payload()
        self.assertTrue("validation_type" in p)
        self.assertTrue("cv_method" in p)
        self.assertTrue("partition_key_cols" in p)
        self.assertEqual(p["validation_type"], "CV")
        self.assertIsNone(p["partition_key_cols"])

    def test_datetime_collect_payload_unsupervised_mode(self):
        # if unsupervised_mode is not set or False we don't want it in payload
        spec = DatetimePartitioningSpecification("date", unsupervised_mode=False)
        payload = spec.collect_payload()
        assert "unsupervised_mode" not in payload

        # otherwise it should be present
        spec = DatetimePartitioningSpecification("date", unsupervised_mode=True)
        payload = spec.collect_payload()
        assert payload["unsupervised_mode"] is True


def test_eligiblity_result_repr():
    er = EligibilityResult(False)
    representation = str(er)
    assert "supported=False" in representation
    assert "reason=''" in representation
    assert "context=''" in representation

    er = EligibilityResult(True, reason="reason", context="context")
    representation = str(er)
    assert "supported=True" in representation
    assert "reason='reason'" in representation
    assert "context='context'" in representation


def test_eligibility_result_with_japanese(unicode_string):
    er = EligibilityResult(True, reason=unicode_string, context=unicode_string)
    print(er)  # test __repr__



import json

import pytest
import responses

from datarobot import Model
from datarobot.errors import ClientError
from datarobot.models.confusion_chart import ConfusionChart


@pytest.fixture
def confusion_chart_validation_data(model_id, confusion_chart_data):
    return {"source": "validation", "sourceModelId": model_id, "data": confusion_chart_data}


@pytest.fixture
def confusion_chart_parent_model_validation_data(parent_model_id, confusion_chart_data):
    return {"source": "validation", "sourceModelId": parent_model_id, "data": confusion_chart_data}


@pytest.fixture
def confusion_chart_parent_model_holdout_data(parent_model_id, confusion_chart_data):
    return {"source": "holdout", "sourceModelId": parent_model_id, "data": confusion_chart_data}


@pytest.fixture
def confusion_chart_validation_metadata_server_data():
    return {
        "source": "validation",
        "classNames": ["1", "2", "3"],
        "totalMatrixSum": 24,
        "relevantClassesPositions": [[1]],
    }


@pytest.fixture
def confusion_chart_holdout_metadata_server_data():
    return {
        "source": "holdout",
        "classNames": ["1", "2", "3"],
        "totalMatrixSum": 24,
        "relevantClassesPositions": [[1]],
    }


def test_instantiation(confusion_chart_validation_data):
    confusion_chart_validation_data["data"]["classes"] = ["1", "2", "3"]
    confusion_chart = ConfusionChart.from_server_data(confusion_chart_validation_data)

    assert confusion_chart.classes == ["1", "2", "3"]
    assert confusion_chart.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert confusion_chart.source == "validation"
    assert confusion_chart.source_model_id == confusion_chart_validation_data["sourceModelId"]


def test_future_proof(confusion_chart_validation_data):
    data_with_future_keys = dict(confusion_chart_validation_data, new_key="some future data")
    data_with_future_keys["data"]["classMetrics"][0]["new_key"] = "some future bin data"
    data_with_future_keys["data"]["classes"] = ["1", "2", "3"]
    confusion_chart = ConfusionChart.from_server_data(data_with_future_keys)

    assert confusion_chart.classes == ["1", "2", "3"]
    assert confusion_chart.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert confusion_chart.source == "validation"


@pytest.fixture
def confusion_chart_validation_data_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/confusionCharts/{}/".format(
        project_id, model_id, "validation"
    )


@pytest.fixture
def confusion_chart_parent_model_validation_data_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/confusionCharts/{}/".format(
        project_id, parent_model_id, "validation"
    )


@pytest.fixture
def confusion_chart_list_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/confusionCharts/".format(
        project_id, model_id
    )


@pytest.fixture
def confusion_chart_parent_list_url(project_id, parent_model_id):
    return "https://host_name.com/projects/{}/models/{}/confusionCharts/".format(
        project_id, parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return "https://host_name.com/projects/{}/frozenModels/{}/".format(project_id, model_id)


@pytest.fixture
def confusion_chart_validation_metadata_response(
    project_url, model_id, confusion_chart_validation_metadata_server_data
):
    responses.add(
        responses.GET,
        "{}models/{}/confusionCharts/validation/metadata/".format(project_url, model_id),
        json=confusion_chart_validation_metadata_server_data,
        content_type="application/json",
    )


@pytest.fixture
def confusion_chart_parent_model_validation_metadata_response(
    project_url, parent_model_id, confusion_chart_validation_metadata_server_data
):
    responses.add(
        responses.GET,
        "{}models/{}/confusionCharts/validation/metadata/".format(project_url, parent_model_id),
        json=confusion_chart_validation_metadata_server_data,
        content_type="application/json",
    )


@pytest.fixture
def confusion_chart_parent_model_holdout_metadata_response(
    project_url, parent_model_id, confusion_chart_validation_metadata_server_data
):
    responses.add(
        responses.GET,
        "{}models/{}/confusionCharts/holdout/metadata/".format(project_url, parent_model_id),
        json=confusion_chart_validation_metadata_server_data,
        content_type="application/json",
    )


@responses.activate
@pytest.mark.usefixtures("confusion_chart_validation_metadata_response")
def test_get_validation_confusion_chart(
    confusion_chart_validation_data, confusion_chart_validation_data_url, project_id, model_id
):
    responses.add(
        responses.GET,
        confusion_chart_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(confusion_chart_validation_data),
    )
    model = Model(id=model_id, project_id=project_id)
    cm = model.get_confusion_chart("validation")

    assert cm.source == confusion_chart_validation_data["source"]
    assert cm.classes == ["1", "2", "3"]
    assert cm.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm.source_model_id == model_id


@responses.activate
def test_get_frozen_validation_confusion_chart_no_fallback(
    confusion_chart_validation_data_url, project_id, model_id
):
    responses.add(responses.GET, confusion_chart_validation_data_url, status=404)
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_confusion_chart("validation")


@responses.activate
@pytest.mark.usefixtures(
    "confusion_chart_parent_model_validation_metadata_response", "known_warning"
)
def test_get_frozen_validation_confusion_chart_with_fallback(
    confusion_chart_parent_model_validation_data,
    confusion_chart_validation_data_url,
    confusion_chart_parent_model_validation_data_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):

    responses.add(responses.GET, confusion_chart_validation_data_url, status=404)
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        confusion_chart_parent_model_validation_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(confusion_chart_parent_model_validation_data),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    cm = model.get_confusion_chart("validation", fallback_to_parent_insights=True)

    assert cm.source == confusion_chart_parent_model_validation_data["source"]
    assert cm.classes == ["1", "2", "3"]
    assert cm.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm.source_model_id == parent_model_id


@responses.activate
@pytest.mark.usefixtures("confusion_chart_validation_metadata_response")
def test_get_all_confusion_chart(
    confusion_chart_validation_data, confusion_chart_list_url, project_id, model_id
):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [confusion_chart_validation_data]}),
    )
    model = Model(id=model_id, project_id=project_id)
    cm_list = model.get_all_confusion_charts()

    assert len(cm_list) == 1
    assert cm_list[0].source == confusion_chart_validation_data["source"]
    assert cm_list[0].classes == ["1", "2", "3"]
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures("confusion_chart_validation_metadata_response")
def test_get_frozen_all_confusion_chart_no_fallback(
    confusion_chart_validation_data, confusion_chart_list_url, project_id, model_id
):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [confusion_chart_validation_data]}),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    cm_list = model.get_all_confusion_charts()

    assert len(cm_list) == 1
    assert cm_list[0].source == confusion_chart_validation_data["source"]
    assert cm_list[0].classes == ["1", "2", "3"]
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures(
    "confusion_chart_validation_metadata_response",
    "confusion_chart_parent_model_validation_metadata_response",
    "confusion_chart_parent_model_holdout_metadata_response",
    "known_warning",
)
def test_get_frozen_all_confusion_chart_with_fallback(
    confusion_chart_validation_data,
    confusion_chart_parent_model_validation_data,
    confusion_chart_parent_model_holdout_data,
    confusion_chart_list_url,
    confusion_chart_parent_list_url,
    frozen_model_url,
    project_id,
    model_id,
    parent_model_id,
    frozen_json,
):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"charts": [confusion_chart_validation_data]}),
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        confusion_chart_parent_list_url,
        status=200,
        content_type="application/json",
        body=json.dumps(
            {
                "charts": [
                    confusion_chart_parent_model_validation_data,
                    confusion_chart_parent_model_holdout_data,
                ]
            }
        ),
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    cm_list = model.get_all_confusion_charts(fallback_to_parent_insights=True)

    assert len(cm_list) == 2
    assert cm_list[0].source == confusion_chart_validation_data["source"]
    assert cm_list[0].classes == ["1", "2", "3"]
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[0].source_model_id == model_id

    assert cm_list[1].source == confusion_chart_parent_model_holdout_data["source"]
    assert cm_list[1].classes == ["1", "2", "3"]
    assert cm_list[1].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert cm_list[1].source_model_id == parent_model_id



import copy
import json
import os

import contextlib2
from requests_toolbelt import MultipartEncoder
from six import string_types
import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT, NETWORK_EGRESS_POLICY
from datarobot.models.api_object import APIObject
from datarobot.models.job import filter_feature_impact_result
from datarobot.models.validators import feature_impact_trafaret
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution, wait_for_custom_resolution


class CustomModelFileItem(APIObject):
    """A file item attached to a DataRobot custom model version.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        id of the file item
    file_name: str
        name of the file item
    file_path: str
        path of the file item
    file_source: str
        source of the file item
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("file_name"): t.String(),
            t.Key("file_path"): t.String(),
            t.Key("file_source"): t.String(),
            t.Key("created", optional=True) >> "created_at": t.String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self, id, file_name, file_path, file_source, created_at=None,
    ):
        self.id = id
        self.file_name = file_name
        self.file_path = file_path
        self.file_source = file_source
        self.created_at = created_at


class CustomModelVersionDependencyBuild(APIObject):
    """Metadata about a DataRobot custom model version's dependency build

    .. versionadded:: v2.22

    Attributes
    ----------
    custom_model_id: str
        id of the custom model
    custom_model_version_id: str
        id of the custom model version
    build_status: str
        the status of the custom model version's dependency build
    started_at: str
        ISO-8601 formatted timestamp of when the build was started
    completed_at: str, optional
        ISO-8601 formatted timestamp of when the build has completed
    """

    _path = "customModels/{}/versions/{}/dependencyBuild/"
    _log_path = "customModels/{}/versions/{}/dependencyBuildLog/"

    _converter = t.Dict(
        {
            t.Key("custom_model_id"): t.String(),
            t.Key("custom_model_version_id"): t.String(),
            t.Key("build_status"): t.String(),
            t.Key("build_start") >> "started_at": t.String(),
            t.Key("build_end", optional=True) >> "completed_at": t.String(allow_blank=True),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}(model={!r}, version={!r}, status={!r})".format(
                self.__class__.__name__,
                self.custom_model_id,
                self.custom_model_version_id,
                self.build_status,
            )
        )

    def _set_values(
        self, custom_model_id, custom_model_version_id, build_status, started_at, completed_at=None,
    ):
        self.custom_model_id = custom_model_id
        self.custom_model_version_id = custom_model_version_id
        self.build_status = build_status
        self.started_at = started_at
        self.completed_at = completed_at

    @classmethod
    def _update_server_data(cls, server_data, custom_model_id, custom_model_version_id):
        updated_data = copy.copy(server_data)
        updated_data.update(
            {"customModelId": custom_model_id, "customModelVersionId": custom_model_version_id}
        )
        return updated_data

    @classmethod
    def get_build_info(cls, custom_model_id, custom_model_version_id):
        """Retrieve information about a custom model version's dependency build

        .. versionadded:: v2.22

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version

        Returns
        -------
        CustomModelVersionDependencyBuild
            the dependency build information
        """
        url = cls._path.format(custom_model_id, custom_model_version_id)
        response = cls._client.get(url)
        server_data = response.json()
        updated_data = cls._update_server_data(
            server_data, custom_model_id, custom_model_version_id
        )
        return cls.from_server_data(updated_data)

    @classmethod
    def start_build(cls, custom_model_id, custom_model_version_id, max_wait=DEFAULT_MAX_WAIT):
        """Start the dependency build for a custom model version  dependency build

        .. versionadded:: v2.22

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version
        max_wait: int, optional
            max time to wait for a build completion.
            If set to None - method will return without waiting.
        """

        def build_complete(response):
            data = response.json()
            if data["buildStatus"] in ["success", "failed"]:
                updated_data = cls._update_server_data(
                    data, custom_model_id, custom_model_version_id
                )
                return cls.from_server_data(updated_data)
            return None

        url = cls._path.format(custom_model_id, custom_model_version_id)
        response = cls._client.post(url)

        if max_wait is None:
            server_data = response.json()
            updated_data = cls._update_server_data(
                server_data, custom_model_id, custom_model_version_id
            )
            return cls.from_server_data(updated_data)
        else:
            return wait_for_custom_resolution(cls._client, url, build_complete, max_wait)

    def get_log(self):
        """Get log of a custom model version dependency build.

        .. versionadded:: v2.22

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._log_path.format(self.custom_model_id, self.custom_model_version_id)
        return self._client.get(url).text

    def cancel(self):
        """Cancel custom model version dependency build that is in progress.

        .. versionadded:: v2.22

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.custom_model_id, self.custom_model_version_id)
        self._client.delete(url)

    def refresh(self):
        """Update custom model version dependency build with the latest data from server.

        .. versionadded:: v2.22

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.custom_model_id, self.custom_model_version_id)

        response = self._client.get(url)

        data = response.json()
        updated_data = self._update_server_data(
            data, self.custom_model_id, self.custom_model_version_id
        )
        self._set_values(**self._safe_data(updated_data, do_recursive=True))


class CustomDependencyConstraint(APIObject):
    """Metadata about a constraint on a dependency of a custom model version

    .. versionadded:: v2.22

    Attributes
    ----------
    constraint_type: str
        How the dependency should be constrained by version (<, <=, ==, >=, >)
    version: str
        The version to use in the dependency's constraint
    """

    _converter = t.Dict(
        {t.Key("constraint_type"): t.String(), t.Key("version"): t.String()}
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}(constraint_type={!r}, version={!r})".format(
                self.__class__.__name__, self.constraint_type, self.version,
            )
        )

    def _set_values(self, constraint_type, version):
        self.constraint_type = constraint_type
        self.version = version


class CustomDependency(APIObject):
    """Metadata about an individual dependency of a custom model version

    .. versionadded:: v2.22

    Attributes
    ----------
    package_name: str
        The dependency's package name
    constraints: List[CustomDependencyConstraint]
        Version constraints to apply on the dependency
    line: str
        The original line from the requirements file
    line_number: int
        The line number the requirement was on in the requirements file
    """

    _converter = t.Dict(
        {
            t.Key("package_name"): t.String(),
            t.Key("constraints"): t.List(CustomDependencyConstraint.schema),
            t.Key("line"): t.String(),
            t.Key("line_number"): t.Int(gt=0),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}(package_name={!r}, constraints={!r})".format(
                self.__class__.__name__, self.package_name, self.constraints,
            )
        )

    def _set_values(self, package_name, constraints, line, line_number):
        self.package_name = package_name
        self.constraints = [CustomDependencyConstraint(**c) for c in constraints]
        self.line = line
        self.line_number = line_number


class CustomModelVersion(APIObject):
    """A version of a DataRobot custom model.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        id of the custom model version
    custom_model_id: str
        id of the custom model
    version_minor: int
        a minor version number of custom model version
    version_major: int
        a major version number of custom model version
    is_frozen: bool
        a flag if the custom model version is frozen
    items: List[CustomModelFileItem]
        a list of file items attached to the custom model version
    base_environment_id: str
        id of the environment to use with the model
    base_environment_version_id: str
        id of the environment version to use with the model
    label: str, optional
        short human readable string to label the version
    description: str, optional
        custom model version description
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created
    dependencies: List[CustomDependency]
        the parsed dependencies of the custom model version if the
        version has a valid requirements.txt file
    network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
        Determines whether the given custom model is isolated, or can access the public network.
        Can be either 'datarobot.NONE' or 'datarobot.PUBLIC'
    maximum_memory: int, optional
        The maximum memory that might be allocated by the custom-model.
        If exceeded, the custom-model will be killed by k8s
    replicas: int, optional
        A fixed number of replicas that will be deployed in the cluster
    required_metadata: Dict[str, str]
        Additional parameters required by the execution environment. The required keys are
        defined by the fieldNames in the base envionment's requiredMetadataKeys. Once set,
        they cannot be changed. If you to change them, make a new version.
    """

    _path = "customModels/{}/versions/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("custom_model_id"): t.String(),
            t.Key("version_minor"): t.Int(),
            t.Key("version_major"): t.Int(),
            t.Key("is_frozen"): t.Bool(),
            t.Key("items"): t.List(CustomModelFileItem.schema),
            # base_environment_id will be required once dependency management is enabled by default
            # in 6.2, but for backwards compatibility, this should be optional
            t.Key("base_environment_id", optional=True): t.String(),
            t.Key("base_environment_version_id", optional=True): t.String(),
            t.Key("label", optional=True): t.String(max_length=50, allow_blank=True) | t.Null(),
            t.Key("description", optional=True): t.String(max_length=10000, allow_blank=True)
            | t.Null(),
            t.Key("created", optional=True) >> "created_at": t.String(),
            t.Key("dependencies", optional=True): t.List(CustomDependency.schema),
            t.Key("network_egress_policy", optional=True): t.Enum(*NETWORK_EGRESS_POLICY.ALL),
            t.Key("maximum_memory", optional=True): t.Int(),
            t.Key("replicas", optional=True): t.Int(),
            t.Key("required_metadata", optional=True): t.Mapping(t.String(), t.String()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({!r})".format(self.__class__.__name__, self.label or self.id)
        )

    def _set_values(
        self,
        id,
        custom_model_id,
        version_minor,
        version_major,
        is_frozen,
        items,
        base_environment_id=None,
        base_environment_version_id=None,
        label=None,
        description=None,
        created_at=None,
        dependencies=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
        required_metadata=None,
    ):
        self.id = id
        self.custom_model_id = custom_model_id
        self.version_minor = version_minor
        self.version_major = version_major
        self.is_frozen = is_frozen
        self.items = [CustomModelFileItem(**item) for item in items]
        self.base_environment_id = base_environment_id
        self.base_environment_version_id = base_environment_version_id
        self.label = label
        self.description = description
        self.created_at = created_at
        self.dependencies = [CustomDependency(**dep) for dep in dependencies or []]
        self.network_egress_policy = network_egress_policy
        self.maximum_memory = maximum_memory
        self.replicas = replicas
        self.required_metadata = required_metadata

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        initial = super(CustomModelVersion, cls).from_server_data(data, keep_attrs)
        # from_server_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        initial.required_metadata = data.get("requiredMetadata")
        return initial

    @classmethod
    def create_clean(
        cls,
        custom_model_id,
        base_environment_id,
        is_major_update=True,
        folder_path=None,
        files=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
        required_metadata=None,
    ):
        """Create a custom model version without files from previous versions.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        base_environment_id: str
            the id of the base environment to use with the custom model version
        is_major_update: bool
            the flag defining if a custom model version
            will be a minor or a major version.
            Default to `True`
        folder_path: str, optional
            the path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path
        files: list, optional
            the list of tuples, where values in each tuple are the local filesystem path and
            the path the file should be placed in the model.
            if list is of strings, then basenames will be used for tuples
            Example:
            [("/home/user/Documents/myModel/file1.txt", "file1.txt"),
            ("/home/user/Documents/myModel/folder/file2.txt", "folder/file2.txt")]
            or
            ["/home/user/Documents/myModel/file1.txt",
            "/home/user/Documents/myModel/folder/file2.txt"]
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Can be either 'datarobot.NONE' or 'datarobot.PUBLIC'
        maximum_memory: int, optional
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s
        replicas: int, optional
            A fixed number of replicas that will be deployed in the cluster
        required_metadata: Dict[str, str]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base envionment's requiredMetadataKeys. Once set,
            they cannot be changed. If you to change them, make a new version.

        Returns
        -------
        CustomModelVersion
            created custom model version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if files and not isinstance(files[0], tuple) and isinstance(files[0], string_types):
            files = [(filename, os.path.basename(filename)) for filename in files]
        return cls._create(
            "post",
            custom_model_id,
            is_major_update,
            base_environment_id,
            folder_path,
            files,
            network_egress_policy=network_egress_policy,
            maximum_memory=maximum_memory,
            replicas=replicas,
            required_metadata=required_metadata,
        )

    @classmethod
    def create_from_previous(
        cls,
        custom_model_id,
        base_environment_id,
        is_major_update=True,
        folder_path=None,
        files=None,
        files_to_delete=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
        required_metadata=None,
    ):
        """Create a custom model version containing files from a previous version.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        base_environment_id: str
            the id of the base environment to use with the custom model version
        is_major_update: bool, optional
            the flag defining if a custom model version
            will be a minor or a major version.
            Default to `True`
        folder_path: str, optional
            the path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path
        files: list, optional
            the list of tuples, where values in each tuple are the local filesystem path and
            the path the file should be placed in the model.
            if list is of strings, then basenames will be used for tuples
            Example:
            [("/home/user/Documents/myModel/file1.txt", "file1.txt"),
            ("/home/user/Documents/myModel/folder/file2.txt", "folder/file2.txt")]
            or
            ["/home/user/Documents/myModel/file1.txt",
            "/home/user/Documents/myModel/folder/file2.txt"]
        files_to_delete: list, optional
            the list of a file items ids to be deleted
            Example: ["5ea95f7a4024030aba48e4f9", "5ea6b5da402403181895cc51"]
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Can be either 'datarobot.NONE' or 'datarobot.PUBLIC'
        maximum_memory: int, optional
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s
        replicas: int, optional
            A fixed number of replicas that will be deployed in the cluster
        required_metadata: Dict[str, str]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base envionment's requiredMetadataKeys. Once set,
            they cannot be changed. If you to change them, make a new version.

        Returns
        -------
        CustomModelVersion
            created custom model version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if files and not isinstance(files[0], tuple) and type(files[0]) in string_types:
            files = [(filename, os.path.basename(filename)) for filename in files]
        if files_to_delete:
            upload_data = [("filesToDelete", file_id) for file_id in files_to_delete]
        else:
            upload_data = None
        return cls._create(
            "patch",
            custom_model_id,
            is_major_update,
            base_environment_id,
            folder_path,
            files,
            upload_data,
            network_egress_policy,
            maximum_memory,
            replicas,
            required_metadata=required_metadata,
        )

    @classmethod
    def _create(
        cls,
        method,
        custom_model_id,
        is_major_update,
        base_environment_id,
        folder_path=None,
        files=None,
        extra_upload_data=None,
        network_egress_policy=None,
        maximum_memory=None,
        replicas=None,
        required_metadata=None,
    ):
        url = cls._path.format(custom_model_id)

        with contextlib2.ExitStack() as stack:
            upload_data = [
                ("isMajorUpdate", str(is_major_update)),
                ("baseEnvironmentId", base_environment_id),
            ]

            if folder_path:
                for root_path, _, file_paths in os.walk(folder_path):
                    for path in file_paths:
                        file_path = os.path.join(root_path, path)
                        file = stack.enter_context(open(file_path, "rb"))

                        upload_data.append(("file", (os.path.basename(file_path), file)))
                        upload_data.append(("filePath", os.path.relpath(file_path, folder_path)))

            if files:
                for file_path, upload_file_path in files:
                    file = stack.enter_context(open(file_path, "rb"))

                    upload_data.append(("file", (os.path.basename(upload_file_path), file)))
                    upload_data.append(("filePath", upload_file_path))

            if extra_upload_data:
                upload_data += extra_upload_data

            if network_egress_policy:
                upload_data.append(("networkEgressPolicy", network_egress_policy))

            if maximum_memory:
                upload_data.append(("maximumMemory", str(maximum_memory)))

            if replicas:
                upload_data.append(("replicas", str(replicas)))

            if required_metadata:
                upload_data.append(("requiredMetadata", json.dumps(required_metadata)))

            encoder = MultipartEncoder(fields=upload_data)
            headers = {"Content-Type": encoder.content_type}
            response = cls._client.request(method, url, data=encoder, headers=headers)
            return cls.from_server_data(response.json())

    @classmethod
    def list(cls, custom_model_id):
        """List custom model versions.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model

        Returns
        -------
        List[CustomModelVersion]
            a list of custom model versions

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._path.format(custom_model_id)
        data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_model_id, custom_model_version_id):
        """Get custom model version by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version to retrieve

        Returns
        -------
        CustomModelVersion
            retrieved custom model version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = cls._path.format(custom_model_id)
        path = "{}{}/".format(url, custom_model_version_id)
        return cls.from_location(path)

    def download(self, file_path):
        """Download custom model version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            path to create a file with custom model version content

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id)
        path = "{}{}/download/".format(url, self.id)

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    def update(self, description=None, required_metadata=None):
        """Update custom model version properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        description: str
            new custom model version description
        required_metadata: Dict[str, str]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base envionment's requiredMetadataKeys. Once set,
            they cannot be changed. If you to change them, make a new version.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {}
        if description:
            payload.update({"description": description})
        if required_metadata:
            payload.update({"requiredMetadata": required_metadata})

        url = self._path.format(self.custom_model_id)
        path = "{}{}/".format(url, self.id)

        response = self._client.patch(path, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))
        # _safe_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        self.required_metadata = data.get("requiredMetadata")

    def refresh(self):
        """Update custom model version with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.custom_model_id)
        path = "{}{}/".format(url, self.id)

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def get_feature_impact(self, with_metadata=False):
        """Get custom model feature impact.

        .. versionadded:: v2.23

        Parameters
        ----------
        with_metadata : bool
            The flag indicating if the result should include the metadata as well.

        Returns
        -------
        feature_impacts : list of dict
           The feature impact data. Each item is a dict with the keys 'featureName',
           'impactNormalized', and 'impactUnnormalized', and 'redundantWith'.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id)
        path = "{}{}/featureImpact/".format(url, self.id)
        data = self._client.get(path).json()
        data = feature_impact_trafaret.check(data)
        return filter_feature_impact_result(data, with_metadata=with_metadata)

    def calculate_feature_impact(self, max_wait=DEFAULT_MAX_WAIT):
        """Calculate custom model feature impact.

        .. versionadded:: v2.23

        Parameters
        ----------
        max_wait: int, optional
            max time to wait for feature impact calculation.
            If set to None - method will return without waiting.
            Defaults to 10 min

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.custom_model_id)
        path = "{}{}/featureImpact/".format(url, self.id)
        response = self._client.post(path)

        if max_wait is not None:
            wait_for_async_resolution(self._client, response.headers["Location"], max_wait)



# -*- encoding: utf-8 -*-

import json
import unittest

import mock
import pandas as pd
import pytest
import responses

from datarobot import errors, Job, PredictJob, Project
from datarobot.models.predict_job import wait_for_async_predictions
from datarobot.utils import raw_prediction_response_to_dataframe
from tests.test_helpers import fixture_file_path
from tests.utils import assert_raised_regex, SDKTestcase

SAMPLE_BINARY_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Binary",
        u"predictions": [
            {
                u"positiveProbability": 0.9,
                u"prediction": 1.0,
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.9, u"label": u"1.0"},
                    {u"value": 0.1, u"label": u"0.0"},
                ],
            },
            {
                u"positiveProbability": 0.1,
                u"prediction": 0.0,
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.9, u"label": u"1.0"},
                    {u"value": 0.1, u"label": u"0.0"},
                ],
            },
            {
                u"positiveProbability": 0.8,
                u"prediction": 1.0,
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.8, u"label": u"1.0"},
                    {u"value": 0.2, u"label": u"0.0"},
                ],
            },
        ],
    }
)


SAMPLE_BINARY_FLOAT_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Binary",
        u"predictions": [
            {
                u"positiveProbability": 0.9,
                u"prediction": 1.0,
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.9, u"label": 1.0},
                    {u"value": 0.1, u"label": 0.0},
                ],
            },
            {
                u"positiveProbability": 0.1,
                u"prediction": 0.0,
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.9, u"label": 1.0},
                    {u"value": 0.1, u"label": 0.0},
                ],
            },
            {
                u"positiveProbability": 0.8,
                u"prediction": 1.0,
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.8, u"label": 1.0},
                    {u"value": 0.2, u"label": 0.0},
                ],
            },
        ],
    }
)


SAMPLE_MULTICLASS_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Multiclass",
        u"predictions": [
            {
                u"positiveProbability": None,
                u"prediction": u"class_a",
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.7, u"label": u"class_a"},
                    {u"value": 0.2, u"label": u"class_b"},
                    {u"value": 0.1, u"label": u"class_c"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_a",
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.6, u"label": u"class_a"},
                    {u"value": 0.2, u"label": u"class_b"},
                    {u"value": 0.2, u"label": u"class_c"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_b",
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.1, u"label": u"class_a"},
                    {u"value": 0.8, u"label": u"class_b"},
                    {u"value": 0.1, u"label": u"class_c"},
                ],
            },
        ],
    }
)


UNICODE_SAMPLE_MULTICLASS_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Multiclass",
        u"predictions": [
            {
                u"positiveProbability": None,
                u"prediction": u"class_赤木",
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.7, u"label": u"class_赤木"},
                    {u"value": 0.2, u"label": u"class_エピソード"},
                    {u"value": 0.1, u"label": u"class_外"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_赤木",
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.6, u"label": u"class_赤木"},
                    {u"value": 0.2, u"label": u"class_エピソード"},
                    {u"value": 0.2, u"label": u"class_外"},
                ],
            },
            {
                u"positiveProbability": None,
                u"prediction": u"class_エピソード",
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.1, u"label": u"class_赤木"},
                    {u"value": 0.8, u"label": u"class_エピソード"},
                    {u"value": 0.1, u"label": u"class_外"},
                ],
            },
        ],
    }
)


SAMPLE_REGRESSION_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": None,
        u"task": u"Regression",
        u"predictions": [
            {u"positiveProbability": None, u"prediction": 32.0, u"rowId": 0},
            {u"positiveProbability": None, u"prediction": 100.0, u"rowId": 1},
            {u"positiveProbability": None, u"prediction": 212.0, u"rowId": 2},
        ],
    }
)


SAMPLE_REGRESSION_PREDICTION_WITH_SHAP_EXPL_RETURN = json.dumps(
    {
        "shapBaseValue": [31108.28125],
        "task": "Regression",
        "positiveClass": None,
        "predictions": [
            {
                "positiveProbability": None,
                "predictionExplanationMetadata": {"shapRemainingTotal": -6969.941378326501},
                "prediction": 27229.166666666668,
                "rowId": 0,
                "predictionExplanations": [
                    {
                        "featureValue": "2004",
                        "strength": 3090.8262156678165,
                        "feature": "YearMade",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "999089",
                        "strength": -2788.4580943655205,
                        "feature": "MachineID",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "2006",
                        "strength": -2286.181442643504,
                        "feature": "saledate (Year)",
                        "label": "SalePrice",
                    },
                ],
            },
            {
                "positiveProbability": None,
                "predictionExplanationMetadata": {"shapRemainingTotal": 13439.981135022792},
                "prediction": 60583.333333333336,
                "rowId": 1,
                "predictionExplanations": [
                    {
                        "featureValue": "117657",
                        "strength": 16035.069717943523,
                        "feature": "MachineID",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "77",
                        "strength": 10105.427194858292,
                        "feature": "ModelID",
                        "label": "SalePrice",
                    },
                    {
                        "featureValue": "Medium",
                        "strength": 3432.899846833262,
                        "feature": "ProductSize",
                        "label": "SalePrice",
                    },
                ],
            },
        ],
    }
)


class TestPredictJob(SDKTestcase):
    def setUp(self):
        super(TestPredictJob, self).setUp()
        self.project_id = "556902e8100d2b3728d47551"
        self.model_id = "556902ef100d2b37da13077d"
        self.predict_job_id = 111
        self.post_url = "https://host_name.com/projects/{}/predictions/".format(self.project_id)
        self.get_job_url = "https://host_name.com/projects/{}/predictJobs/{}/".format(
            self.project_id, self.predict_job_id
        )
        self.get_predictions_url = "https://host_name.com/projects/{}/predictions/123/".format(
            self.project_id
        )
        self.predict_job_data = {
            "status": "queue",
            "id": self.predict_job_id,
            "projectId": self.project_id,
            "modelId": self.model_id,
            "isBlocked": False,
        }

        self.predict_job_with_message = {
            "status": "error",
            "id": self.predict_job_id,
            "projectId": self.project_id,
            "modelId": self.model_id,
            "isBlocked": False,
            "message": "Clearly this is user error :)",
        }

        self.mock_generic_job_data = {
            "status": "inprogress",
            "url": "https://host_name.com/projects/{}/predictJobs/1/".format(self.project_id),
            "id": "1",
            "jobType": "predict",
            "isBlocked": False,
            "projectId": self.project_id,
        }

        self.mock_model_job_data = {
            "status": "inprogress",
            "url": "https://host_name.com/projects/{}/modelJobs/2/".format(self.project_id),
            "id": "2",
            "jobType": "model",
            "isBlocked": False,
            "projectId": self.project_id,
        }

    def test_instantiate_predict_job(self):
        job = PredictJob(self.predict_job_data)
        print(job)  # checks that method relevant to print works on this (failed in one release)
        self.assertEqual(job.id, self.predict_job_id)
        self.assertEqual(job.status, "queue")
        self.assertIsInstance(job.project, Project)
        self.assertEqual(job.project.id, "556902e8100d2b3728d47551")

        with self.assertRaises(ValueError):
            PredictJob("qid")

    def test_future_proof(self):
        PredictJob(dict(self.predict_job_data, future="new"))

    @responses.activate
    def test_from_job(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/predictJobs/1/".format(self.project_id),
            status=200,
            body=json.dumps(self.predict_job_with_message),
            content_type="application/json",
        )
        generic_job = Job(self.mock_generic_job_data)
        predict_job = PredictJob.from_job(generic_job)

        self.assertEqual(predict_job.status, "error")
        self.assertEqual(predict_job.job_type, "predict")
        self.assertEqual(predict_job.project, Project(self.project_id))
        self.assertEqual(predict_job.model.id, self.model_id)

    def test_from_job_wrong_type(self):
        generic_job = Job(self.mock_model_job_data)
        with self.assertRaises(ValueError):
            PredictJob.from_job(generic_job)

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_create_non_ascii_filename(self):
        path = fixture_file_path(u"日本/データ.csv")
        model = mock.Mock()
        model.id = self.model_id
        model.project.id = self.project_id
        responses.add(
            responses.POST,
            self.post_url,
            status=202,
            body="",
            adding_headers={"Location": "https://host_name.com" + self.get_job_url},
        )
        predict_job_id = PredictJob.create(model, sourcedata=path)
        assert predict_job_id == str(self.predict_job_id)

    @pytest.mark.usefixtures("known_warning")
    def test_create_bad_sourcedata(self):
        with pytest.raises(errors.InputNotUnderstoodError) as exc_info:
            PredictJob.create(mock.Mock(), sourcedata=123)
        assert_raised_regex(exc_info, "sourcedata parameter not understood.")

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_create_success(self):
        model = mock.Mock()
        model.id = self.model_id
        model.project.id = self.project_id
        responses.add(
            responses.POST,
            self.post_url,
            status=202,
            body="",
            adding_headers={"Location": "https://host_name.com" + self.get_job_url},
        )
        predict_job_id = PredictJob.create(model, b"one,two\n3,4")
        self.assertEqual(predict_job_id, str(self.predict_job_id))

    @responses.activate
    def test_cancel(self):
        predict_job = PredictJob(self.predict_job_data)
        responses.add(
            responses.DELETE, self.get_job_url, status=204, body="", content_type="application/json"
        )
        predict_job.cancel()
        self.assertEqual(responses.calls[0].request.method, "DELETE")

    @responses.activate
    def test_get_unfinished_predict_job_with_message(self):
        """A new attribute of PredictJob beginning in DataRobot API v2.1"""
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_with_message),
            content_type="application/json",
        )
        predict_job = PredictJob.get(self.project_id, self.predict_job_id)
        self.assertIsInstance(predict_job, PredictJob)
        self.assertEqual(predict_job.id, self.predict_job_id)
        self.assertEqual(predict_job.status, "error")
        self.assertEqual(predict_job.message, "Clearly this is user error :)")

    @responses.activate
    def test_get_unfinished_predict_job(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_data),
            content_type="application/json",
        )
        predict_job = PredictJob.get(self.project_id, self.predict_job_id)
        self.assertIsInstance(predict_job, PredictJob)
        self.assertEqual(predict_job.id, self.predict_job_id)
        self.assertEqual(predict_job.status, "queue")
        self.assertEqual(predict_job.message, "")

    @responses.activate
    def test_get_finished_predict_job(self):
        responses.add(
            responses.GET, self.get_job_url, status=303, body="",
        )
        with self.assertRaises(errors.PendingJobFinished):
            PredictJob.get(self.project_id, self.predict_job_id)

    @responses.activate
    def test_get_predict_job_async_failure(self):
        responses.add(
            responses.GET, self.get_job_url, status=202, body="",
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            PredictJob.get(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code 202")

    @responses.activate
    def test_get_unfinished_predictions(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_data),
            content_type="application/json",
        )
        with pytest.raises(errors.JobNotFinished) as exc_info:
            PredictJob.get_predictions(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, "Pending job status is queue")

    @responses.activate
    def test_get_predictions_async_failure(self):
        responses.add(
            responses.GET, self.get_job_url, status=202, body="",
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            PredictJob.get_predictions(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code 202")

    @responses.activate
    def test_get_finished_predictions(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=303,
            body="",
            adding_headers={"Location": self.get_predictions_url},
        )
        responses.add(
            responses.GET,
            self.get_predictions_url,
            status=200,
            body=SAMPLE_BINARY_PREDICTION_RETURN,
            content_type="application/json",
        )
        predictions = PredictJob.get_predictions(self.project_id, self.predict_job_id)
        self.assertIsInstance(predictions, pd.DataFrame)
        self.assertEqual(predictions.shape, (3, 5))
        self.assertEqual(
            sorted(predictions.columns),
            sorted(["positive_probability", "prediction", "row_id", "class_0.0", "class_1.0"]),
        )


class TestWaitForAsyncPredictions(SDKTestcase):
    def setUp(self):
        super(TestWaitForAsyncPredictions, self).setUp()
        self.pid = "p-id"
        self.predict_job_id = "5"
        self.get_job_url = "https://host_name.com/projects/{}/predictJobs/{}/".format(
            self.pid, self.predict_job_id,
        )

    @responses.activate
    def test_timeout(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps({"isBlocked": False}),
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncTimeoutError) as exc_info:
            wait_for_async_predictions(self.pid, self.predict_job_id, max_wait=1)
        assert_raised_regex(exc_info, "Predictions generation timed out in")

    @responses.activate
    @mock.patch("datarobot.models.predict_job.PredictJob.get_predictions")
    def test_success(self, get_predictions):
        responses.add(
            responses.GET, self.get_job_url, status=303, body="", content_type="application/json"
        )
        wait_for_async_predictions(self.pid, self.predict_job_id)
        self.assertEqual(get_predictions.call_count, 1)

    @responses.activate
    @mock.patch("datarobot.models.predict_job.PredictJob.get_predictions")
    def test_error(self, get_predictions):
        data = {"status": "error", "isBlocked": False}
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(data),
            content_type="application/json",
        )
        with pytest.raises(errors.AsyncPredictionsGenerationError) as exc_info:
            wait_for_async_predictions(self.pid, self.predict_job_id)
        assert_raised_regex(exc_info, "Predictions generation unsuccessful")


class TestRawPredictionResponseToDataframe(unittest.TestCase):
    def test_parse_regression_predictions(self):
        data = json.loads(SAMPLE_REGRESSION_PREDICTION_RETURN)

        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 2))
        self.assertEqual(
            sorted(frame.columns), [u"prediction", u"row_id"],
        )

    def test_parse_classification_predictions(self):
        data = json.loads(SAMPLE_BINARY_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"positive_probability", u"prediction", u"row_id", u"0.0", u"1.0"]),
        )

    def test_parse_classification_predictions_with_floats(self):
        data = json.loads(SAMPLE_BINARY_FLOAT_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "class_")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"positive_probability", u"prediction", u"row_id", u"class_0.0", u"class_1.0"]),
        )

    def test_parse_multiclass_classification_predictions(self):
        data = json.loads(SAMPLE_MULTICLASS_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"prediction", u"row_id", u"class_a", u"class_b", u"class_c"]),
        )

    def test_parse_multiclass_classification_predictions_unicode(self):
        data = json.loads(UNICODE_SAMPLE_MULTICLASS_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data, "")
        self.assertEqual(frame.shape, (3, 5))
        self.assertEqual(
            sorted(frame.columns),
            sorted([u"prediction", u"row_id", u"class_赤木", u"class_エピソード", u"class_外"]),
        )

    def test_parse_regression_predictions_with_shap_explanations(self):
        data = json.loads(SAMPLE_REGRESSION_PREDICTION_WITH_SHAP_EXPL_RETURN)
        frame = raw_prediction_response_to_dataframe(data, class_prefix="")
        assert frame.shape == (2, 13)
        assert sorted(frame.columns) == sorted(
            [
                "prediction",
                "row_id",
                "Explanation_1_feature_name",
                "Explanation_1_feature_value",
                "Explanation_1_strength",
                "Explanation_2_feature_name",
                "Explanation_2_feature_value",
                "Explanation_2_strength",
                "Explanation_3_feature_name",
                "Explanation_3_feature_value",
                "Explanation_3_strength",
                "shap_remaining_total",
                "shap_base_value",
            ]
        )

        # assert scalar shapBaseValue doesn't cause an error
        data["shapBaseValue"] = data["shapBaseValue"][0]
        frame_2 = raw_prediction_response_to_dataframe(data, class_prefix="")
        assert frame.equals(frame_2)

        # when shapBaseValue has multiple items the first one is used and no error is raised
        data["shapBaseValue"] = [9.12, 42.1, -27.1]
        frame_3 = raw_prediction_response_to_dataframe(data, class_prefix="")
        assert frame_3["shap_base_value"].tolist() == [9.12, 9.12]



import copy
from copy import deepcopy
import json

import mock
import pytest
import responses

from datarobot import CustomModelTest


@pytest.fixture
def custom_model_id():
    return "5cf4d3f5f930e26daac18acc"


@pytest.fixture
def mocked_tests():
    return {
        "count": 1,
        "totalCount": 1,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18acc",
                "datasetId": "5e60d80b6eb24502c8814ed2",
                "datasetVersionId": "5d31ffd2417a06005178268b",
                "customModelImageId": "5e9ec6503129820040debd8c",
                "imageType": "customModelVersion",
                "overallStatus": "in_progress",
                "testingStatus": {
                    "error_check": {"status": "in_progress", "message": "msg"},
                    "null_value_imputation": {"status": "in_progress", "message": "msg"},
                    "long_running_service": {"status": "in_progress", "message": "msg"},
                    "side_effects": {"status": "in_progress", "message": "msg"},
                },
                "completedAt": "2019-09-28T15:19:26.587583Z",
                "createdBy": "5e60d80b6eb24502c8814ed2",
                "created": "2019-09-28T15:19:26.587583Z",
                "networkEgressPolicy": None,
                "maximumMemory": None,
                "replicas": None,
            },
        ],
    }


@pytest.fixture
def mocked_tests_with_resources(mocked_tests):
    tests = copy.deepcopy(mocked_tests)
    tests["data"][0].update(
        {"networkEgressPolicy": "PUBLIC", "maximumMemory": 1024 * 1024 * 1024, "replicas": 3}
    )
    return tests


@pytest.fixture
def mocked_test(mocked_tests):
    return mocked_tests["data"][0]


@pytest.fixture
def mocked_test_future_test(mocked_tests):
    test = copy.deepcopy(mocked_tests["data"][0])
    test["testingStatus"]["future_test"] = {"status": "in_progress", "message": "msg"}
    return test


@pytest.fixture
def mocked_test_future_field_in_status(mocked_tests):
    test = copy.deepcopy(mocked_tests["data"][0])
    test["testingStatus"]["error_check"]["future_field"] = "chrome"
    return test


@pytest.fixture
def make_tests_url(unittest_endpoint):
    def _make_tests_url(test_id=None, custom_model_id=None):
        url = "{}/customModelTests/".format(unittest_endpoint)
        if test_id is not None:
            url = "{}{}/".format(url, test_id)
        if custom_model_id is not None:
            url += "?customModelId={}".format(custom_model_id)
        return url

    return _make_tests_url


def mock_get_response(url, response):
    responses.add(
        responses.GET, url, status=200, content_type="application/json", body=json.dumps(response),
    )


def assert_test(test, test_json):
    assert test.id == test_json["id"]
    assert test.dataset_id == test_json["datasetId"]
    assert test.dataset_version_id == test_json["datasetVersionId"]
    assert test.custom_model_image_id == test_json["customModelImageId"]
    assert test.image_type == test_json["imageType"]
    assert test.overall_status == test_json["overallStatus"]
    assert test.detailed_status == test_json["testingStatus"]
    assert test.created_by == test_json["createdBy"]
    assert test.completed_at == test_json["completedAt"]
    assert test.created_at == test_json["created"]


def test_from_server_data(mocked_test):
    test = CustomModelTest.from_server_data(mocked_test)
    assert_test(test, mocked_test)


@responses.activate
@pytest.mark.parametrize("include_env", [True, False])
@pytest.mark.parametrize("mocked_fixture_name", ["mocked_tests", "mocked_tests_with_resources"])
def test_create_non_blocking(
    request,
    unittest_endpoint,
    mocked_fixture_name,
    make_tests_url,
    custom_model_id,
    tmpdir,
    include_env,
):
    mocked_tests = request.getfixturevalue(mocked_fixture_name)
    status_url = "{}/status_url".format(unittest_endpoint)

    responses.add(
        responses.POST,
        make_tests_url(),
        status=202,
        content_type="application/json",
        headers={"Location": status_url},
    )
    responses.add(
        responses.GET,
        make_tests_url(custom_model_id=custom_model_id),
        status=200,
        content_type="application/json",
        headers={"Location": status_url},
        body=json.dumps(mocked_tests),
    )

    args = [
        custom_model_id,
        "custom model version id",
        "dataset id",
    ]
    if include_env:
        args += ["environment id", "environment version id"]

    data = mocked_tests["data"][0]

    test = CustomModelTest.create(
        *args,
        max_wait=None,
        network_egress_policy=data["networkEgressPolicy"],
        maximum_memory=data["maximumMemory"],
        replicas=data["replicas"]
    )

    assert_test(test, mocked_tests["data"][0])

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_tests_url()
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == make_tests_url(custom_model_id=custom_model_id)

    request_body = json.loads(responses.calls[0].request.body)
    if include_env:
        assert all(field in request_body for field in ["environmentId", "environmentVersionId"])
    else:
        assert not any(field in request_body for field in ["environmentId", "environmentVersionId"])


@responses.activate
def test_create_blocking(unittest_endpoint, mocked_tests, make_tests_url, custom_model_id, tmpdir):
    status_url = "{}/status_url".format(unittest_endpoint)

    test = mocked_tests["data"][0]
    test_url = make_tests_url(test["id"])

    responses.add(
        responses.POST,
        make_tests_url(),
        status=202,
        content_type="application/json",
        headers={"Location": status_url},
    )
    responses.add(
        responses.GET,
        make_tests_url(custom_model_id=custom_model_id),
        status=200,
        content_type="application/json",
        headers={"Location": status_url},
        body=json.dumps(mocked_tests),
    )
    responses.add(
        responses.GET,
        status_url,
        status=303,
        content_type="application/json",
        headers={"Location": test_url},
    )
    responses.add(
        responses.GET, test_url, status=200, content_type="application/json", body=json.dumps(test),
    )

    test_created = CustomModelTest.create(
        custom_model_id,
        "custom model version id",
        "dataset id",
        "environment id",
        "environment version id",
    )

    assert_test(test_created, test)

    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_tests_url()
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == make_tests_url(custom_model_id=custom_model_id)
    assert responses.calls[2].request.method == "GET"
    assert responses.calls[2].request.url == status_url
    assert responses.calls[3].request.method == "GET"
    assert responses.calls[3].request.url == test_url


@responses.activate
@mock.patch("time.sleep", return_value=None)
def test_create_blocking_error(
    sleep_mock, unittest_endpoint, mocked_tests, make_tests_url, custom_model_id, tmpdir
):
    status_url = "{}/status_url".format(unittest_endpoint)

    test = mocked_tests["data"][0]
    test_url = make_tests_url(test["id"])

    responses.add(
        responses.POST,
        make_tests_url(),
        status=202,
        content_type="application/json",
        headers={"Location": status_url},
    )
    responses.add(
        responses.GET,
        make_tests_url(custom_model_id=custom_model_id),
        status=200,
        content_type="application/json",
        headers={"Location": status_url},
        body=json.dumps(mocked_tests),
    )
    responses.add(
        responses.GET,
        status_url,
        status=200,
        content_type="application/json",
        body=json.dumps({"status": "abort"}),
    )
    test["overallStatus"] = "in_progress"
    responses.add(
        responses.GET, test_url, status=200, content_type="application/json", body=json.dumps(test),
    )
    test["overallStatus"] = "failed"
    responses.add(
        responses.GET, test_url, status=200, content_type="application/json", body=json.dumps(test),
    )

    test_created = CustomModelTest.create(
        custom_model_id,
        "custom model version id",
        "dataset id",
        "environment id",
        "environment version id",
    )

    assert_test(test_created, test)

    sleep_mock.call_count == 1
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == make_tests_url()
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == make_tests_url(custom_model_id=custom_model_id)
    assert responses.calls[2].request.method == "GET"
    assert responses.calls[2].request.url == status_url
    assert responses.calls[3].request.method == "GET"
    assert responses.calls[3].request.url == test_url
    assert json.loads(responses.calls[3].response.content)["overallStatus"] == "in_progress"
    assert responses.calls[4].request.url == test_url
    assert json.loads(responses.calls[4].response.content)["overallStatus"] == "failed"


@responses.activate
@pytest.mark.parametrize(
    "test_json", ["mocked_test", "mocked_test_future_test", "mocked_test_future_field_in_status"],
)
def test_get_test(request, test_json, make_tests_url):
    mocked_test = request.getfixturevalue(test_json)
    url = make_tests_url(mocked_test["id"])
    mock_get_response(url, mocked_test)

    test = CustomModelTest.get(mocked_test["id"])
    assert_test(test, mocked_test)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_tests(mocked_tests, make_tests_url, custom_model_id):
    url = make_tests_url(custom_model_id=custom_model_id)
    mock_get_response(url, mocked_tests)

    tests = CustomModelTest.list(custom_model_id)

    assert len(tests) == len(mocked_tests["data"])
    for test, mocked_test in zip(tests, mocked_tests["data"]):
        assert_test(test, mocked_test)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_tests_multiple_pages(mocked_tests, make_tests_url, custom_model_id):
    url1 = make_tests_url(custom_model_id=custom_model_id)
    url2 = make_tests_url(custom_model_id=custom_model_id) + "2"

    mocked_tests_2nd = deepcopy(mocked_tests)
    mocked_tests["next"] = url2

    mock_get_response(url1, mocked_tests)
    mock_get_response(url2, mocked_tests_2nd)

    tests = CustomModelTest.list(custom_model_id)
    assert len(tests) == len(mocked_tests["data"]) + len(mocked_tests_2nd["data"])
    for test, mocked_test in zip(tests, mocked_tests["data"] + mocked_tests_2nd["data"]):
        assert_test(test, mocked_test)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_get_log(mocked_test, make_tests_url):
    url = make_tests_url(mocked_test["id"])
    log_url = url + "log/"

    mock_get_response(url, mocked_test)
    responses.add(responses.GET, log_url, status=200, content_type="text/plain", body="log content")

    test = CustomModelTest.get(mocked_test["id"])
    log = test.get_log()

    assert log == "log content"

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == log_url


@responses.activate
def test_get_log_tail(mocked_test, make_tests_url):
    url = make_tests_url(mocked_test["id"])
    log_url = url + "tail/"

    mock_get_response(url, mocked_test)
    responses.add(responses.GET, log_url, status=200, content_type="text/plain", body="log content")

    test = CustomModelTest.get(mocked_test["id"])
    log = test.get_log_tail()

    assert log == "log content"

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == log_url


@responses.activate
def test_cancel(mocked_test, make_tests_url):
    url = make_tests_url(mocked_test["id"])

    mock_get_response(url, mocked_test)
    responses.add(
        responses.DELETE, url, status=200,
    )

    test = CustomModelTest.get(mocked_test["id"])
    test.cancel()

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "DELETE"
    assert responses.calls[1].request.url == url


@responses.activate
def test_refresh(mocked_test, make_tests_url):
    url = make_tests_url(mocked_test["id"])

    mock_get_response(url, mocked_test)

    mocked_test.update({"overallStatus": "failed"})

    mock_get_response(url, mocked_test)

    test = CustomModelTest.get(mocked_test["id"])
    test.refresh()
    assert_test(test, mocked_test)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url == url



import copy
import json
import os

import pytest
import responses

from datarobot import Model
from datarobot._experimental.models.custom_task import CustomTask
from datarobot.enums import CUSTOM_MODEL_TARGET_TYPE, CUSTOM_TASK_TYPE
from datarobot.errors import ClientError
from datarobot.utils import underscorize
from tests.model.test_custom_model import assert_custom_model_common, mock_get_response


@pytest.fixture
def mocked_custom_estimator(mocked_models):
    return mocked_models["data"][0]


@pytest.fixture
def mocked_custom_transform(mocked_models):
    models = copy.deepcopy(mocked_models)
    models["data"][0].update({"targetType": "Transform"})

    return models["data"][0]


@pytest.fixture
def model_artifact_contents():
    # stolen from scoring code test
    return "some_random_file_contents" * 1000000


@pytest.fixture
def local_artifact_filename(tmpdir):
    return str(tmpdir.join("temp.zip"))


@pytest.mark.parametrize("custom_mock", ["mocked_custom_estimator", "mocked_custom_transform"])
class TestCustomTask(object):
    def test_from_server_data(self, custom_mock, request):
        task_dict = request.getfixturevalue(custom_mock)
        task = CustomTask.from_server_data(task_dict)
        assert_custom_model_common(task, task_dict, custom_task=True)

    @responses.activate
    def test_get_version(self, request, custom_mock, make_tasks_url):
        # arrange
        mocked_task = request.getfixturevalue(custom_mock)
        url = make_tasks_url(mocked_task["id"])
        mock_get_response(url, mocked_task)

        # act
        task = CustomTask.get(mocked_task["id"])

        # assert
        assert_custom_model_common(task, mocked_task, custom_task=True)
        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    def test_list_tasks(self, request, custom_mock, make_tasks_url):
        mocked_task = request.getfixturevalue(custom_mock)
        url = make_tasks_url()
        mock_get_response(url, {"data": [mocked_task], "next": None})

        response = CustomTask.list()

        assert len(response) == 1
        assert_custom_model_common(response[0], mocked_task, custom_task=True)
        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url

    @responses.activate
    def test_update(self, request, custom_mock, make_tasks_url):
        # arrange
        patch = {
            "description": "new description",
        }
        mocked_task = request.getfixturevalue(custom_mock)
        url = make_tasks_url(mocked_task["id"])
        mock_get_response(url, mocked_task)

        mocked_task.update(patch)

        responses.add(
            responses.PATCH,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_task),
        )

        # act
        model = CustomTask.get(mocked_task["id"])
        args = copy.deepcopy(patch)

        model.update(**{underscorize(k): v for k, v in args.items()})

        # assert
        assert_custom_model_common(model, mocked_task, custom_task=True)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url == url
        assert responses.calls[1].request.body == json.dumps(patch).encode()

    @responses.activate
    def test_refresh(self, request, custom_mock, make_tasks_url):
        # arrange
        mocked_task = request.getfixturevalue(custom_mock)
        url = make_tasks_url(mocked_task["id"])
        mock_get_response(url, mocked_task)

        mocked_task.update({"description": "new desc"})

        mock_get_response(url, mocked_task)

        # act
        model = CustomTask.get(mocked_task["id"])
        model.refresh()

        # assert
        assert_custom_model_common(model, mocked_task, custom_task=True)

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_delete(self, request, custom_mock, make_tasks_url):
        mocked_task = request.getfixturevalue(custom_mock)
        url = make_tasks_url(mocked_task["id"])

        mock_get_response(url, mocked_task)
        responses.add(
            responses.DELETE, url, status=204, content_type="application/json",
        )

        model = CustomTask.get(mocked_task["id"])
        model.delete()

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "DELETE"
        assert responses.calls[1].request.url == url

    @responses.activate
    def test_download(self, request, custom_mock, make_tasks_url, tmpdir):
        # arrange
        mocked_task = request.getfixturevalue(custom_mock)
        url = make_tasks_url(mocked_task["id"])
        mock_get_response(url, mocked_task)
        responses.add(
            responses.GET,
            url + "download/",
            status=200,
            content_type="application/json",
            body=b"content",
        )

        downloaded_file = tmpdir.mkdir("sub").join("download")
        downloaded_file_path = str(downloaded_file)

        # act
        model = CustomTask.get(mocked_task["id"])
        model.download_latest_version(downloaded_file_path)

        # assert
        downloaded_file.read() == b"content"

        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == url
        assert responses.calls[1].request.method == "GET"
        assert responses.calls[1].request.url == url + "download/"

    @responses.activate
    def test_create(self, request, custom_mock, make_tasks_url):
        mocked_task = request.getfixturevalue(custom_mock)

        responses.add(
            responses.POST,
            make_tasks_url(),
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_task),
        )

        if mocked_task["targetType"] == CUSTOM_TASK_TYPE.TRANSFORM:
            target_type = None
            task_type = CUSTOM_TASK_TYPE.TRANSFORM
        else:
            target_type = CUSTOM_MODEL_TARGET_TYPE.REGRESSION
            task_type = CUSTOM_TASK_TYPE.ESTIMATOR

        model = CustomTask.create(
            name=mocked_task["name"],
            task_type=task_type,
            target_type=target_type,
            description=mocked_task["description"],
        )

        assert_custom_model_common(model, mocked_task, custom_task=True)

        assert responses.calls[0].request.method == "POST"
        assert responses.calls[0].request.url == make_tasks_url()

    @responses.activate
    def test_create_copy(self, request, custom_mock, make_tasks_url):
        mocked_task = request.getfixturevalue(custom_mock)

        custom_task_to_copy = mocked_task["id"]

        url = make_tasks_url() + "fromCustomTask/"
        responses.add(
            responses.POST,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(mocked_task),
        )

        model = CustomTask.copy_custom_task(custom_task_to_copy)
        assert_custom_model_common(model, mocked_task, custom_task=True)

        req = responses.calls[0].request
        assert req.method == "POST"
        assert req.url == url
        assert json.loads(req.body) == {"customTaskId": custom_task_to_copy}


class TestArtifactDownload(object):
    @responses.activate
    def test_positive_case(
        self, project_id, model_id, model_artifact_contents, local_artifact_filename
    ):
        url = "https://host_name.com/projects/{}/models/{}/trainingArtifact/".format(
            project_id, model_id
        )
        responses.add(
            responses.GET, url, body=model_artifact_contents, status=200, match_querystring=True
        )
        model = Model(project_id=project_id, id=model_id)
        model.download_training_artifact(local_artifact_filename)
        assert os.path.isfile(local_artifact_filename)
        with open(local_artifact_filename) as f:
            contents = f.read()
            assert contents == model_artifact_contents
        os.unlink(local_artifact_filename)
        assert not os.path.isfile(local_artifact_filename)
        req = responses.calls[0].request
        assert req.method == "GET"
        assert req.url == url

    @responses.activate
    def test_error(self, project_id, model_id, model_artifact_contents, local_artifact_filename):
        url = "https://host_name.com/projects/{}/models/{}/trainingArtifact/".format(
            project_id, model_id
        )
        error_message = "some error"
        responses.add(
            responses.GET, url, status=404, match_querystring=True, json={"message": error_message},
        )
        model = Model(project_id=project_id, id=model_id)
        with pytest.raises(ClientError) as exc_info:
            model.download_training_artifact(local_artifact_filename)

        assert not os.path.isfile(local_artifact_filename)
        assert error_message in str(exc_info.value)



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import get_id_from_response


class Ruleset(APIObject):
    """ Represents an approximation of a model with DataRobot Prime

    Attributes
    ----------
    id : str
        the id of the ruleset
    rule_count : int
        the number of rules used to approximate the model
    score : float
        the validation score of the approximation
    project_id : str
        the project the approximation belongs to
    parent_model_id : str
        the model being approximated
    model_id : str or None
        the model using this ruleset (if it exists).  Will be None if no such model has been
        trained.

    """

    _converter = t.Dict(
        {
            t.Key("project_id"): t.String(),
            t.Key("parent_model_id"): t.String(),
            t.Key("model_id", optional=True): t.String(),
            t.Key("ruleset_id"): t.Int(),
            t.Key("rule_count"): t.Int(),
            t.Key("score"): t.Float(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        project_id=None,
        parent_model_id=None,
        model_id=None,
        ruleset_id=None,
        rule_count=None,
        score=None,
    ):
        self.id = ruleset_id
        self.rule_count = rule_count
        self.score = score
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id

    def __repr__(self):
        return "Ruleset(rule_count={}, score={})".format(self.rule_count, self.score)

    def request_model(self):
        """ Request training for a model using this ruleset

        Training a model using a ruleset is a necessary prerequisite for being able to download
        the code for a ruleset.

        Returns
        -------
        job: Job
            the job fitting the new Prime model
        """
        from . import Job

        if self.model_id:
            raise ValueError("Model already exists for ruleset")
        url = "projects/{}/primeModels/".format(self.project_id)
        data = {"parent_model_id": self.parent_model_id, "ruleset_id": self.id}
        response = self._client.post(url, data=data)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)
